import csv
from collections import defaultdict
from copy import copy
from enum import Enum, auto
from functools import lru_cache
from itertools import chain, product
from math import isnan
from typing import Any, DefaultDict, Dict, Iterable, Mapping, Optional, Sequence

import numpy as np
from tqdm import tqdm

from .annotation import Annotation
from .annotationset import AnnotationSet
from .atomic import open_atomic
from .boundingbox import BoundingBox
from .file_utils import PathLike
from .utils import all_equal, grouping, mean


class RecallSteps(Enum):
    ELEVEN = auto()
    ALL = auto()


class PartialEvaluationItem:
    def __init__(
        self,
        tps: Optional["list[bool]"] = None,
        scores: Optional["list[float]"] = None,
        npos: int = 0,
    ) -> None:
        self._tps = tps or []
        self._scores = scores or []
        self._npos = npos
        self._cache: dict[str, Any] = {}

        assert self._npos >= 0, f"'npos' ({npos}) should be greater than or equal to 0."
        assert (
            len(self._tps) == len(self._scores)
        ), f"The number of 'tps' ({len(self._tps)}) should be equal to the number of 'scores' ({len(self._scores)})."

    def __iadd__(self, other: "PartialEvaluationItem") -> "PartialEvaluationItem":
        self._tps += other._tps
        self._scores += other._scores
        self._npos += other._npos
        self.clear_cache()
        return self

    def __add__(self, other: "PartialEvaluationItem") -> "PartialEvaluationItem":
        copy_ = copy(self)
        copy_ += other
        return copy_

    @property
    def ndet(self) -> int:
        return len(self._tps)

    @property
    def npos(self) -> int:
        return self._npos

    def tp(self) -> int:
        tp = self._cache.get("tp", sum(self._tps))
        self._cache["tp"] = tp
        assert tp <= self._npos
        return tp

    def ap(self) -> float:
        ap = self._cache.get("ap")
        if ap is not None:
            return ap
        ap = COCOEvaluator._compute_ap(self._scores, self._tps, self._npos)
        self._cache["ap"] = ap
        return ap

    def ar(self) -> Optional[float]:
        ar = self._cache.get("ar")
        if ar is not None:
            return ar
        tp = self.tp()
        ar = tp / self._npos if self._npos != 0 else float("nan")
        self._cache["ar"] = ar
        return ar

    def clear_cache(self):
        self._cache.clear()

    def evaluate(self) -> "EvaluationItem":
        return EvaluationItem(self.tp(), self.ndet, self._npos, self.ap(), self.ar())


class EvaluationItem:
    """Evaluation of COCO metrics for one label."""

    __slots__ = ("tp", "ndet", "npos", "ap", "ar")

    def __init__(
        self, tp: int, ndet: int, npos: int, ap: Optional[float], ar: Optional[float]
    ) -> None:
        self.tp = tp
        self.ndet = ndet
        self.npos = npos
        self.ap = ap
        self.ar = ar


class PartialEvaluation(DefaultDict[str, PartialEvaluationItem]):
    """Do not mutate this excepted with defined methods."""

    def __init__(self, items: Optional[Mapping[str, PartialEvaluationItem]] = None):
        super().__init__(lambda: PartialEvaluationItem())
        if items is not None:
            self.update(items)
        self._cache = {}

    def __iadd__(self, other: "PartialEvaluation") -> "PartialEvaluation":
        for key, value in other.items():
            self[key] += value
        self.clear_cache()
        return self

    def __add__(self, other: "PartialEvaluation") -> "PartialEvaluation":
        copy_ = copy(self)
        copy_ += other
        return copy_

    def ap(self) -> float:
        ap = self._cache.get("ap")
        if ap is not None:
            return ap
        ap = mean(a for a in (ev.ap() for ev in self.values()) if not isnan(a))
        self._cache["ap"] = ap
        return ap

    def ar(self) -> Optional[float]:
        ar = self._cache.get("ar")
        if ar is not None:
            return ar
        ar = mean(a for a in (ev.ar() for ev in self.values()) if not isnan(a))
        self._cache["ar"] = ar
        return ar

    def clear_cache(self):
        self._cache.clear()

    def evaluate(self) -> "Evaluation":
        return Evaluation(self)


class Evaluation(Dict[str, EvaluationItem]):
    """Evaluation of COCO metrics for multiple labels."""

    def __init__(self, evaluation: PartialEvaluation) -> None:
        super().__init__()
        self.update({label: ev.evaluate() for label, ev in evaluation.items()})

        self._ap = evaluation.ap()
        self._ar = evaluation.ar()

    def ap(self) -> float:
        return self._ap

    def ar(self) -> float:
        return self._ar


class MultiThresholdEvaluation(Dict[str, Dict[str, float]]):
    """
    Evaluation of COCO metrics for multiple labels and multiple
    IoU thresholds.
    """

    def __init__(self, evaluations: "list[Evaluation]") -> None:
        result: "defaultdict[str, list[EvaluationItem]]" = defaultdict(list)

        for evaluation in evaluations:
            for label, ev_item in evaluation.items():
                result[label].append(ev_item)

        super().__init__(
            {
                label: {
                    "ap": mean(ev.ap for ev in evs if not isnan(ev.ap)),
                    "ar": mean(ev.ar for ev in evs if not isnan(ev.ar)),
                }
                for label, evs in result.items()
            }
        )

    def ap(self) -> float:
        return mean(ev["ap"] for ev in self.values() if not isnan(ev["ap"]))

    def ar(self) -> float:
        return mean(ev["ar"] for ev in self.values() if not isnan(ev["ar"]))


class COCOEvaluator:
    """
    COCO metrics evaluator.

    Once instantiated, the standard COCO metrics can be queried using the dedicated methods:

    * `COCOEvaluator.ap()`,
    * `COCOEvaluator.ap_50()`,
    * `COCOEvaluator.ar_medium()`,
    * etc.

    Custom evaluations can be performed with `COCOEvaluator.evaluate()` and the standard
    COCO metrics can be printed to the console with `COCOEvaluator.show_summary()`.
    """

    AP_THRESHOLDS = np.linspace(0.5, 0.95, 10)
    """The ten COCO AP thresholds: [0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9, 0.95]."""

    SMALL_RANGE = (0.0, 32.0**2)
    """The COCO 'small range' (0, 1024) of bounding box areas (in pixels)."""

    MEDIUM_RANGE = (32.0**2, 96.0**2)
    """The COCO 'medium range' (1024, 9216) of bounding box areas (in pixels)."""

    LARGE_RANGE = (96.0**2, float("inf"))
    """The COCO 'large range' (9216, inf) of bounding box areas (in pixels)."""

    ALL_RANGE = (0.0, float("inf"))
    """The range (0, inf) of any bounding box areas possible (in pixels)."""

    RECALL_STEPS = np.linspace(0.0, 1.0, 101)
    """The 101 COCO recall steps used to compute the AP score."""

    CSV_HEADERS = (
        "label",
        "AP 50:95",
        "AP 50",
        "AP 75",
        "AP S",
        "AP M",
        "AP L",
        "AR 1",
        "AR 10",
        "AR 100",
        "AR S",
        "AR M",
        "AR L",
    )

    def __init__(
        self,
        *,
        ground_truths: AnnotationSet,
        predictions: AnnotationSet,
        labels: Optional[Iterable[str]] = None,
    ) -> None:
        """
        Intantiate a `COCOEvaluator` with the to-be-evaluated dataset and the ground truth
        reference one.

        All the bounding box annotations from the prediction dataset must have a confidence score
        and all the ones from the ground truth one must not.

        If labels are provided, only those will be evaluated, else every label will.

        The evaluated datasets must not be modified during the lifetime of the `COCOEvaluator` or
        the results will be wrong.
        """
        self._predictions = predictions
        self._ground_truths = ground_truths

        if labels is None:
            self.labels = []
        else:
            self.labels = list(labels)

        self._cached_evaluate = lru_cache(maxsize=60 * 4)(self.__evaluate)

    def evaluate(
        self,
        *,
        iou_threshold: float,
        max_detections: int = 100,
        size_range: Optional["tuple[float, float]"] = None,
    ) -> Evaluation:
        """
        Evaluate COCO metrics for custom parameters.

        Parameters:

        * `iou_threshold`: the bounding box iou threshold to consider a ground-truth to
        detection association valid.
        * `max_detections`: the maximum number of detections taken into account (sorted by
        descreasing confidence).
        * `size_range`: the range of size (bounding box area) to consider.
        """
        return self._cached_evaluate(
            iou_threshold=iou_threshold,
            max_detections=max_detections,
            size_range=size_range,
        )

    def __evaluate(
        self,
        *,
        iou_threshold: float,
        max_detections: int = 100,
        size_range: Optional["tuple[float, float]"] = None,
    ) -> Evaluation:
        if size_range is None:
            size_range = COCOEvaluator.ALL_RANGE

        self._assert_params(iou_threshold, max_detections, size_range)

        return self.evaluate_annotations(
            self._predictions,
            self._ground_truths,
            iou_threshold,
            max_detections,
            size_range,
            self.labels,
        ).evaluate()

    def ap(self) -> float:
        return self.ap_evaluation().ap()

    def ap_50(self) -> float:
        return self.ap_50_evaluation().ap()

    def ap_75(self) -> float:
        return self.ap_75_evaluation().ap()

    def ap_small(self) -> float:
        return self.small_evaluation().ap()

    def ap_medium(self) -> float:
        return self.medium_evaluation().ap()

    def ap_large(self) -> float:
        return self.large_evaluation().ap()

    def ar_1(self) -> float:
        return self.ndets_1_evaluation().ar()

    def ar_10(self) -> float:
        return self.ndets_10_evaluation().ar()

    def ar_100(self) -> float:
        return self.ndets_100_evaluation().ar()

    def ar_small(self) -> float:
        return self.small_evaluation().ar()

    def ar_medium(self) -> float:
        return self.medium_evaluation().ar()

    def ar_large(self) -> float:
        return self.large_evaluation().ar()

    def ap_evaluation(self) -> MultiThresholdEvaluation:
        evaluations = [
            self.evaluate(
                iou_threshold=t, max_detections=100, size_range=self.ALL_RANGE
            )
            for t in self.AP_THRESHOLDS
        ]

        return MultiThresholdEvaluation(evaluations)

    def ap_50_evaluation(self) -> Evaluation:
        return self.evaluate(
            iou_threshold=0.5, max_detections=100, size_range=self.ALL_RANGE
        )

    def ap_75_evaluation(self) -> Evaluation:
        return self.evaluate(
            iou_threshold=0.75, max_detections=100, size_range=self.ALL_RANGE
        )

    def small_evaluation(self) -> MultiThresholdEvaluation:
        return self._range_evalation(self.SMALL_RANGE)

    def medium_evaluation(self) -> MultiThresholdEvaluation:
        return self._range_evalation(self.MEDIUM_RANGE)

    def large_evaluation(self) -> MultiThresholdEvaluation:
        return self._range_evalation(self.LARGE_RANGE)

    def ndets_1_evaluation(self) -> MultiThresholdEvaluation:
        return self._ndets_evaluation(1)

    def ndets_10_evaluation(self) -> MultiThresholdEvaluation:
        return self._ndets_evaluation(10)

    def ndets_100_evaluation(self) -> MultiThresholdEvaluation:
        return self._ndets_evaluation(100)

    def _range_evalation(
        self, range_: "tuple[float, float]"
    ) -> MultiThresholdEvaluation:
        evaluations = [
            self.evaluate(iou_threshold=t, max_detections=100, size_range=range_)
            for t in self.AP_THRESHOLDS
        ]

        return MultiThresholdEvaluation(evaluations)

    def _ndets_evaluation(self, max_dets: int) -> MultiThresholdEvaluation:
        evaluations = [
            self.evaluate(
                iou_threshold=t, max_detections=max_dets, size_range=self.ALL_RANGE
            )
            for t in self.AP_THRESHOLDS
        ]

        return MultiThresholdEvaluation(evaluations)

    @classmethod
    def evaluate_annotations(
        cls,
        predictions: AnnotationSet,
        ground_truths: AnnotationSet,
        iou_threshold: float,
        max_detections: int,
        size_range: "tuple[float, float]",
        labels: Optional[Sequence[str]] = None,
    ) -> PartialEvaluation:
        image_ids = ground_truths.image_ids | predictions.image_ids
        evaluation = PartialEvaluation()

        for image_id in sorted(image_ids):  # Sorted to ensure reproductibility
            gt = ground_truths.get(image_id) or Annotation(image_id)
            pred = predictions.get(image_id) or Annotation(image_id)

            evaluation += cls.evaluate_annotation(
                pred, gt, iou_threshold, max_detections, size_range, labels
            )

        return evaluation

    @classmethod
    def evaluate_annotation(
        cls,
        prediction: Annotation,
        ground_truth: Annotation,
        iou_threshold: float,
        max_detections: int,
        size_range: "tuple[float, float]",
        labels: Optional[Iterable[str]] = None,
    ) -> PartialEvaluation:
        assert (
            prediction.image_id == ground_truth.image_id
        ), f"The prediction image id '{prediction.image_id}' should be the same as the ground truth image id '{ground_truth.image_id}'."

        preds = grouping(prediction.boxes, lambda box: box.label)
        refs = grouping(ground_truth.boxes, lambda box: box.label)
        labels = labels or set(preds.keys()).union(refs.keys())
        evaluation = PartialEvaluation()

        for label in labels:
            dets = preds.get(label, [])
            gts = refs.get(label, [])

            evaluation[label] += cls.evaluate_boxes(
                dets, gts, iou_threshold, max_detections, size_range
            )

        return evaluation

    @classmethod
    def evaluate_boxes(
        cls,
        predictions: "list[BoundingBox]",
        ground_truths: "list[BoundingBox]",
        iou_threshold: float,
        max_detections: int,
        size_range: "tuple[float, float]",
    ) -> PartialEvaluationItem:
        assert all(
            p.is_detection for p in predictions
        ), "Prediction annotations should not contain ground truth annotations."
        assert all(
            g.is_ground_truth for g in ground_truths
        ), "Ground truth annotations should not contain prediction annotations."
        assert all_equal(
            p.label for p in predictions
        ), "The prediction boxes should have the same label."
        assert all_equal(
            g.label for g in ground_truths
        ), "The ground truth boxes should have the same label."

        cls._assert_params(iou_threshold, max_detections, size_range)

        dets = sorted(predictions, key=lambda box: box._confidence, reverse=True)  # type: ignore
        dets = dets[:max_detections]

        gts = sorted(ground_truths, key=lambda box: not box._area_in(size_range))
        gt_ignore = [not g._area_in(size_range) for g in gts]  # Redundant `area_in`

        gt_matches = {}
        dt_matches = {}

        for idx_dt, det in enumerate(dets):
            best_iou = 0.0
            idx_best = -1

            for idx_gt, gt in enumerate(gts):
                if idx_gt in gt_matches:
                    continue
                if idx_best > -1 and not gt_ignore[idx_best] and gt_ignore[idx_gt]:
                    break

                iou = det.iou(gt)
                if iou < best_iou:
                    continue

                best_iou = iou
                idx_best = idx_gt

            if idx_best == -1 or best_iou < iou_threshold:
                continue

            dt_matches[idx_dt] = idx_best
            gt_matches[idx_best] = idx_dt

        # This is overly complicated
        dt_ignore = [
            gt_ignore[dt_matches[i]] if i in dt_matches else not d._area_in(size_range)
            for i, d in enumerate(dets)
        ]

        scores = [d._confidence for i, d in enumerate(dets) if not dt_ignore[i]]
        matches = [i in dt_matches for i in range(len(dets)) if not dt_ignore[i]]
        npos = sum(1 for i in range(len(gts)) if not gt_ignore[i])

        return PartialEvaluationItem(matches, scores, npos)

    @staticmethod
    def _assert_params(
        iou_threshold: float, max_detections: int, size_range: "tuple[float, float]"
    ):
        assert (
            0.0 <= iou_threshold <= 1.0
        ), f"the IoU threshold ({iou_threshold}) should be between 0 and 1."
        assert (
            max_detections >= 0
        ), f"The maximum number of detections ({max_detections}) should be positive."

        low, high = size_range
        assert (
            low >= 0.0 and high >= low
        ), f"The size range '({low}, {high})' should be positive and non-empty, i.e. 0 < size_range[0] <= size_range[1]."

    def clear_cache(self):
        self._cached_evaluate.cache_clear()

    def _evaluate_all(self, *, verbose: bool = False):
        params = chain(
            product(
                self.AP_THRESHOLDS,
                (100,),
                (self.SMALL_RANGE, self.MEDIUM_RANGE, self.LARGE_RANGE, self.ALL_RANGE),
            ),
            product(self.AP_THRESHOLDS, (1, 10), (self.ALL_RANGE,)),
        )

        for t, d, r in tqdm(params, desc="Evaluation", total=60, disable=not verbose):
            self.evaluate(iou_threshold=t, max_detections=d, size_range=r)

    def show_summary(self, *, verbose: bool = False):
        """Compute and show the standard COCO metrics."""
        from rich import print as pprint
        from rich.table import Table

        self._evaluate_all(verbose=verbose)

        table = Table(title="COCO Evaluation", show_footer=True)
        table.add_column("Label", footer="Total")

        metrics = {
            "AP 50:95": self.ap(),
            "AP 50": self.ap_50(),
            "AP 75": self.ap_75(),
            "AP S": self.ap_small(),
            "AP M": self.ap_medium(),
            "AP L": self.ap_large(),
            "AR 1": self.ar_1(),
            "AR 10": self.ar_10(),
            "AR 100": self.ar_100(),
            "AR S": self.ar_small(),
            "AR M": self.ar_medium(),
            "AR L": self.ar_large(),
        }

        for metric_name, metric in metrics.items():
            table.add_column(metric_name, justify="right", footer=f"{metric:.2%}")

        labels = self.labels or sorted(self.ap_evaluation().keys())

        for label in labels:
            ap = self.ap_evaluation()[label]["ap"]
            ap_50 = self.ap_50_evaluation()[label].ap
            ap_75 = self.ap_75_evaluation()[label].ap

            ap_s = self.small_evaluation()[label]["ap"]
            ap_m = self.medium_evaluation()[label]["ap"]
            ap_l = self.large_evaluation()[label]["ap"]

            ar_1 = self.ndets_1_evaluation()[label]["ap"]
            ar_10 = self.ndets_10_evaluation()[label]["ap"]
            ar_100 = self.ndets_100_evaluation()[label]["ap"]

            ar_s = self.small_evaluation()[label]["ar"]
            ar_m = self.medium_evaluation()[label]["ar"]
            ar_l = self.large_evaluation()[label]["ar"]

            table.add_row(
                label,
                f"{ap:.2%}",
                f"{ap_50:.2%}",
                f"{ap_75:.2%}",
                f"{ap_s:.2%}",
                f"{ap_m:.2%}",
                f"{ap_l:.2%}",
                f"{ar_1:.2%}",
                f"{ar_10:.2%}",
                f"{ar_100:.2%}",
                f"{ar_s:.2%}",
                f"{ar_m:.2%}",
                f"{ar_l:.2%}",
            )

        table.header_style = "bold"
        table.footer_style = "bold"
        table.row_styles = ["none", "dim"]

        for c in table.columns[1:4]:
            c.style = "red"
            c.header_style = "red"
            c.footer_style = "red"

        for c in table.columns[4:7]:
            c.style = "magenta"
            c.header_style = "magenta"
            c.footer_style = "magenta"

        for c in table.columns[7:10]:
            c.style = "blue"
            c.header_style = "blue"
            c.footer_style = "blue"

        for c in table.columns[10:13]:
            c.style = "green"
            c.header_style = "green"
            c.footer_style = "green"

        pprint(table)

    def save_csv(self, path: PathLike, *, verbose: bool = False):
        self._evaluate_all(verbose=verbose)
        labels = self.labels or sorted(self.ap_evaluation().keys())

        with open_atomic(path, "w") as f:
            writer = csv.writer(f)
            writer.writerow(COCOEvaluator.CSV_HEADERS)

            for label in labels:
                ap = self.ap_evaluation()[label]["ap"]
                ap_50 = self.ap_50_evaluation()[label].ap
                ap_75 = self.ap_75_evaluation()[label].ap

                ap_s = self.small_evaluation()[label]["ap"]
                ap_m = self.medium_evaluation()[label]["ap"]
                ap_l = self.large_evaluation()[label]["ap"]

                ar_1 = self.ndets_1_evaluation()[label]["ap"]
                ar_10 = self.ndets_10_evaluation()[label]["ap"]
                ar_100 = self.ndets_100_evaluation()[label]["ap"]

                ar_s = self.small_evaluation()[label]["ar"]
                ar_m = self.medium_evaluation()[label]["ar"]
                ar_l = self.large_evaluation()[label]["ar"]

                row = (
                    label,
                    ap,
                    ap_50,
                    ap_75,
                    ap_s,
                    ap_m,
                    ap_l,
                    ar_1,
                    ar_10,
                    ar_100,
                    ar_s,
                    ar_m,
                    ar_l,
                )

                writer.writerow(row)

    @classmethod
    def _compute_ap(
        cls, scores: "list[float]", matched: "list[bool]", NP: int
    ) -> float:
        """
        This curve tracing method has some quirks that do not appear when only unique confidence
        thresholds are used (i.e. Scikit-learn's implementation), however, in order to be
        consistent, the COCO's method is reproduced.

        Copyrights: https://github.com/rafaelpadilla/review_object_detection_metrics
        """
        if NP == 0:
            return float("nan")

        # sort in descending score order
        scores_ = np.array(scores, dtype=float)
        matched_ = np.array(matched, dtype=bool)
        inds = np.argsort(-scores_, kind="stable")

        scores_ = scores_[inds]
        matched_ = matched_[inds]

        tp = np.cumsum(matched_)

        rc = tp / NP
        pr = tp / np.arange(1, matched_.size + 1)
        # make precision monotonically decreasing
        i_pr = np.maximum.accumulate(pr[::-1])[::-1]
        rec_idx = np.searchsorted(rc, cls.RECALL_STEPS, side="left")

        sum_ = i_pr[rec_idx[rec_idx < i_pr.size]].sum()

        return sum_ / rec_idx.size



import argparse
import sys
from pathlib import Path
from typing import Optional

from .annotationset import AnnotationSet
from .boundingbox import BoxFormat
from .evaluation import COCOEvaluator

PARSE_CHOICES = {
    "coco",
    "labelme",
    "pascalvoc",
    "openimage",
    "txt",
    "cvat",
    "yolov5",
    "yolov7",
    "yolo-darknet",
    "via-json",
    "imagenet",
}
PARSE_CHOICES_EXT = {*PARSE_CHOICES, "coco_result"}
SAVE_CHOICES = {*PARSE_CHOICES}


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument("--quiet", "-q", action="store_true")
    parser.add_argument("--threads", "-j", default=None, type=int)

    subparsers = parser.add_subparsers(dest="mode")
    convert_parser = subparsers.add_parser("convert")
    stats_parser = subparsers.add_parser("summary")
    eval_parser = subparsers.add_parser("evaluate")

    add_convert_args(convert_parser)
    add_stats_args(stats_parser)
    add_eval_args(eval_parser)

    return parser.parse_args()


def add_parse_args(
    parser: argparse.ArgumentParser,
    metavar: str = "input",
    label: Optional[str] = None,
):
    parser.add_argument("input", type=Path, metavar=metavar)

    group = parser.add_argument_group("Parse options" if label is None else label)
    group.add_argument(
        "--format", "-f", type=str, choices=PARSE_CHOICES, dest="format_in"
    )
    group.add_argument("--img_folder", "-d", type=Path, default=None)
    group.add_argument("--mapping", "-m", type=Path, default=None, dest="mapping_in")
    group.add_argument(
        "--bb_fmt",
        "-b",
        type=str,
        choices=("ltrb", "ltwh", "xywh"),
        default="ltrb",
        dest="bb_fmt_in",
    )
    group.add_argument(
        "--norm", "-n", type=str, choices=("abs", "rel"), default="abs", dest="norm_in"
    )
    group.add_argument("--ext", "-e", type=str, default=".txt", dest="ext_in")
    group.add_argument("--img_ext", "-g", type=str, default=".jpg", dest="img_ext_in")
    group.add_argument("--sep", "-p", type=str, default=" ", dest="sep_in")


def add_parse_dets_args(parser: argparse.ArgumentParser):
    parser.add_argument("predictions", type=Path)

    group = parser.add_argument_group("Predictions parse options")
    group.add_argument("--format_dets", "-F", type=str, choices=PARSE_CHOICES_EXT)
    group.add_argument("--mapping_dets", "-M", type=Path, default=None)
    group.add_argument(
        "--bb_fmt_dets",
        "-B",
        type=str,
        choices=("ltrb", "ltwh", "xywh"),
        default="ltrb",
        dest="bb_fmt_dets",
    )
    group.add_argument(
        "--norm_dets",
        "-N",
        type=str,
        choices=("abs", "rel"),
        default="abs",
        dest="norm_in_dets",
    )
    group.add_argument("--ext_dets", "-E", type=str, default=".txt")
    group.add_argument("--img_ext_dets", "-G", type=str, default=".jpg")
    group.add_argument("--sep_dets", "-P", type=str, default=" ")


def add_save_args(parser: argparse.ArgumentParser):
    parser.add_argument("output", type=Path)

    group = parser.add_argument_group("Save options")
    group.add_argument(
        "--save_fmt", "-F", type=str, choices=SAVE_CHOICES, dest="format_out"
    )  # TODO: add PARSE_CHOICES_EXT
    group.add_argument(
        "--bb_fmt_out", "-B", type=str, choices=("ltrb", "ltwh", "xywh"), default="ltrb"
    )
    group.add_argument(
        "--norm_out", "-N", type=str, choices=("abs", "rel"), default="abs"
    )
    group.add_argument("--sep_out", "-P", type=str, default=" ")
    group.add_argument("--ext_out", "-E", type=str, default=".txt")
    group.add_argument("--coco_auto_ids", "-A", action="store_true")

    mapping_group = group.add_mutually_exclusive_group()
    mapping_group.add_argument("--mapping_out", "-M", type=Path, default=None)
    mapping_group.add_argument("--reverse_mapping_out", "-R", type=Path, default=None)


def add_stats_args(parser: argparse.ArgumentParser):
    add_parse_args(parser)


def add_convert_args(parser: argparse.ArgumentParser):
    add_parse_args(parser)
    add_save_args(parser)


def add_eval_args(parser: argparse.ArgumentParser):
    add_parse_args(parser, metavar="groundtruths", label="Ground-truths parse options")
    add_parse_dets_args(parser)

    parser.add_argument("--save", "-s", type=Path, default=None, dest="save_csv_path")

    # parser.add_argument("--ap", action="append", dest="metrics")
    # parser.add_argument("--ap50", action="append", dest="metrics")
    # parser.add_argument("--iou", type=int, default=None)  # mutually_exclusive_group()
    # etc...


def parse_annotations(args: argparse.Namespace) -> AnnotationSet:
    input: Path = args.input.expanduser().resolve()
    format_in: str = args.format_in
    verbose: bool = not args.quiet

    if format_in == "coco":
        return AnnotationSet.from_coco(input, verbose=verbose)
    elif format_in == "pascalvoc" or format_in == "imagenet":
        return AnnotationSet.from_xml(input, verbose=verbose)
    elif format_in == "openimage":
        assert (
            args.img_folder is not None
        ), "The image directory must be provided for openimage format (required for reading the image size)."
        img_dir: Path = args.img_folder.expanduser().resolve()
        return AnnotationSet.from_openimage(
            input, image_folder=img_dir, verbose=verbose
        )
    elif format_in == "labelme":
        return AnnotationSet.from_labelme(input, verbose=verbose)
    elif format_in == "cvat":
        return AnnotationSet.from_cvat(input, verbose=verbose)
    elif format_in == "via-json":
        img_dir: Optional[Path] = args.img_folder
        return AnnotationSet.from_via_json(input, image_folder=img_dir)
    else:
        img_ext: str = args.img_ext_in
        image_dir: Optional[Path] = None

        if args.img_folder is not None:
            image_dir = args.img_folder.expanduser().resolve()

        if format_in == "yolo-darknet":
            annotations = AnnotationSet.from_yolo_darknet(
                input, image_folder=image_dir, image_extension=img_ext, verbose=verbose
            )
        elif format_in == "yolov5":
            annotations = AnnotationSet.from_yolo_v5(
                input, image_folder=image_dir, image_extension=img_ext, verbose=verbose
            )
        elif format_in == "yolov7":
            annotations = AnnotationSet.from_yolo_v7(
                input, image_folder=image_dir, image_extension=img_ext, verbose=verbose
            )
        elif format_in == "txt":
            format = BoxFormat.from_string(args.bb_fmt_in)
            relative: bool = args.norm_in == "rel"
            extension: str = args.ext_in
            sep: str = args.sep_in

            annotations = AnnotationSet.from_txt(
                input,
                image_folder=image_dir,
                box_format=format,
                relative=relative,
                file_extension=extension,
                image_extension=img_ext,
                separator=sep,
                verbose=verbose,
            )
        else:
            raise ValueError(f"Input format '{format_in}' unknown")

        if args.mapping_in is not None:
            map_path: Path = args.mapping_in.expanduser().resolve()
            mapping = AnnotationSet.parse_names_file(map_path)
            annotations.map_labels(mapping)

        return annotations


def parse_dets_annotations(
    args: argparse.Namespace,
    coco_gts: Optional[AnnotationSet] = None,
) -> AnnotationSet:
    input: Path = args.predictions.expanduser().resolve()
    format_dets: str = args.format_dets
    verbose: bool = not args.quiet

    if format_dets == "coco":
        return AnnotationSet.from_coco(input, verbose=verbose)
    if format_dets == "coco_result":
        if coco_gts is None:
            raise ValueError(
                "When using 'COCO results', the parsed ground truths must be in 'COCO' format."
            )
        return coco_gts.from_results(input, verbose=verbose)
    elif format_dets == "pascalvoc" or format_dets == "imagenet":
        return AnnotationSet.from_xml(input, verbose=verbose)
    elif format_dets == "openimage":
        assert (
            args.img_folder is not None
        ), "The image directory must be provided for openimage format (required for reading the image size)."
        img_dir: Path = args.img_folder.expanduser().resolve()
        return AnnotationSet.from_openimage(
            input, image_folder=img_dir, verbose=verbose
        )
    elif format_dets == "labelme":
        return AnnotationSet.from_labelme(input, verbose=verbose)
    elif format_dets == "cvat":
        return AnnotationSet.from_cvat(input, verbose=verbose)
    elif format_dets == "via-json":
        img_dir: Optional[Path] = args.img_folder
        return AnnotationSet.from_via_json(input, image_folder=img_dir)
    else:
        img_ext: str = args.img_ext_dets
        image_dir: Optional[Path] = None

        if args.img_folder is not None:
            image_dir = args.img_folder.expanduser().resolve()

        if format_dets == "yolo-darknet":
            annotations = AnnotationSet.from_yolo_darknet(
                input, image_folder=image_dir, image_extension=img_ext, verbose=verbose
            )
        elif format_dets == "yolov5":
            annotations = AnnotationSet.from_yolo_v5(
                input, image_folder=image_dir, image_extension=img_ext, verbose=verbose
            )
        elif format_dets == "yolov7":
            annotations = AnnotationSet.from_yolo_v7(
                input, image_folder=image_dir, image_extension=img_ext, verbose=verbose
            )
        elif format_dets == "txt":
            bb_fmt = BoxFormat.from_string(args.bb_fmt_dets)
            relative: bool = args.norm_in_dets == "rel"
            extension: str = args.ext_dets
            sep: str = args.sep_dets

            annotations = AnnotationSet.from_txt(
                input,
                image_folder=image_dir,
                box_format=bb_fmt,
                relative=relative,
                file_extension=extension,
                image_extension=img_ext,
                separator=sep,
                verbose=verbose,
            )
        else:
            raise ValueError(f"Groundtruth format '{format_dets}' unknown")

        if args.mapping_dets is not None:
            map_path: Path = args.mapping_dets.expanduser().resolve()
            mapping = AnnotationSet.parse_names_file(map_path)
            annotations.map_labels(mapping)

        return annotations


# TODO: Add "coco_result"
def save_annotations(args: argparse.Namespace, annotations: AnnotationSet):
    output: Path = args.output.expanduser().resolve()
    format_out: str = args.format_out
    verbose: bool = not args.quiet

    if args.mapping_out is not None:  # Takes precedence
        map_path: Path = args.mapping_out.expanduser().resolve()
        mapping = AnnotationSet.parse_names_file(map_path)
        annotations.map_labels(mapping)
    elif args.reverse_mapping_out is not None:
        map_path: Path = args.reverse_mapping_out.expanduser().resolve()
        mapping = AnnotationSet.parse_names_file(map_path)
        mapping = {v: k for k, v in mapping.items()}
        annotations.map_labels(mapping)

    if format_out == "coco":
        annotations.save_coco(output, auto_ids=args.coco_auto_ids, verbose=verbose)
    elif format_out == "pascalvoc":
        annotations.save_xml(output, verbose=verbose)
    elif format_out == "openimage":
        annotations.save_openimage(output, verbose=verbose)
    elif format_out == "labelme":
        annotations.save_labelme(output, verbose=verbose)
    elif format_out == "cvat":
        annotations.save_cvat(output, verbose=verbose)
    elif format_out == "via-json":
        image_folder: Optional[Path] = args.img_folder
        assert (
            image_folder is not None
        ), "The image folder must be provided with `--img_folder` for via-json conversion."
        annotations.save_via_json(output, image_folder=image_folder, verbose=verbose)
    else:
        if format_out == "yolo-darknet":
            annotations.save_yolo_darknet(output, verbose=verbose)
        elif format_out == "yolov5":
            annotations.save_yolo_v5(output, verbose=verbose)
        elif format_out == "yolov7":
            annotations.save_yolo_v7(output, verbose=verbose)
        elif format_out == "txt":
            bb_fmt = BoxFormat.from_string(args.bb_fmt_out)
            relative: bool = args.norm_out == "rel"
            extension: str = args.ext_out
            sep: str = args.sep_out

            annotations.save_txt(
                output,
                label_to_id=None,
                box_format=bb_fmt,
                relative=relative,
                separator=sep,
                file_extension=extension,
                verbose=verbose,
            )
        else:
            raise ValueError(f"Save format '{format_out}' unknown")


def evaluate(
    args: argparse.Namespace, groundtruths: AnnotationSet, predictions: AnnotationSet
):
    verbose: bool = not args.quiet

    evaluator = COCOEvaluator(ground_truths=groundtruths, predictions=predictions)

    evaluator.show_summary(verbose=verbose)

    if args.save_csv_path is not None:
        path = args.save_csv_path.expanduser().resolve()
        evaluator.save_csv(path, verbose=False)
        if verbose:
            print(f"Evaluation saved to '{args.save_csv_path}'.", file=sys.stderr)


def main():
    args = parse_args()

    assert (
        args.threads is None or args.threads > 0
    ), f"The number of threads '{args.threads}' should either be None or be greater than 0."

    mode = args.mode
    if mode == "convert":
        annotations = parse_annotations(args)
        save_annotations(args, annotations)
    elif mode == "summary":
        annotations = parse_annotations(args)
        annotations.show_stats()
    elif mode == "evaluate":
        groundtruths = parse_annotations(args)
        coco_gts = groundtruths if args.format_dets == "coco_result" else None
        predictions = parse_dets_annotations(args, coco_gts=coco_gts)
        evaluate(args, groundtruths, predictions)
    else:
        raise ValueError(f"Sub-command '{mode}' not recognized.")


if __name__ == "__main__":
    main()



import atexit
from concurrent.futures import ThreadPoolExecutor
from operator import length_hint
from typing import Callable, Iterable, Optional, TypeVar

from tqdm import tqdm

SHARED_THREAD_POOL = ThreadPoolExecutor()


def at_exit():
    SHARED_THREAD_POOL.shutdown()


atexit.register(at_exit)


U = TypeVar("U")
V = TypeVar("V")


def thread_map(
    fn: Callable[[U], V],
    it: Iterable[U],
    desc: Optional[str] = None,
    total: Optional[int] = None,
    unit: str = "it",
    verbose: bool = False,
) -> "list[V]":
    disable = not verbose
    total = total or length_hint(it)
    results = []
    with tqdm(desc=desc, total=total, unit=unit, disable=disable) as pbar:
        futures = SHARED_THREAD_POOL.map(fn, it)
        for result in futures:
            results.append(result)
            pbar.update()
    return results



import csv
import json
import xml.etree.ElementTree as et
from collections import defaultdict
from functools import partial
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    TypeVar,
    Union,
)
from warnings import warn

from tqdm import tqdm

from .annotation import Annotation
from .atomic import open_atomic
from .boundingbox import BoundingBox, BoxFormat
from .errors import ParsingError, UnknownImageFormat
from .file_utils import PathLike, glob
from .image_utils import IMAGE_EXTENSIONS, get_image_size
from .thread_utils import thread_map

T = TypeVar("T")


class AnnotationSet:
    """
    A set of annotations of multiple and distinct images, most commonly refered to a 'dataset'.
    """

    def __init__(
        self,
        annotations: Optional[Iterable[Annotation]] = None,
        *,
        override=False,
    ):
        """
        Create an `AnnotationSet` from multiple image annotations. Each annotation should be unique,
        i.e. multiple annotations for a single image (as idendified by its `image_id`) is not
        allowed.

        Parameters:

        * `annotation`: an iterable of image annotations.
        * `override`: if `True`, image annotations not unique are allowed and only the last one in
        the iterator will be kept, else an error is thrown.
        """
        # TODO: Add optional addition of labels found during
        # parsing, for instance COCO names and YOLO `.names`.
        # Could also add a (lazy) computed accessor that
        # runs through all boxes to get labels.

        self._annotations: Dict[str, Annotation] = {}

        if annotations is not None:
            for annotation in annotations:
                self.add(annotation, override=override)

        self._id_to_label: Optional["dict[Any, str]"] = None
        self._id_to_imageid: Optional["dict[Any, str]"] = None

    def __getitem__(self, image_id: str) -> Annotation:
        """
        Get the image annotation with the corresponding `image_id`. Will raise an exception
        if the image ID is not present in the dataset.
        """
        return self._annotations[image_id]

    def get(self, image_id: str) -> Optional[Annotation]:
        """
        Get the image annotation with the corresponding `image_id`, if present in the dataset
        (else `None` is returned).
        """
        return self._annotations.get(image_id)

    def __len__(self) -> int:
        """The number of annotations in the dataset."""
        return len(self._annotations)

    def __iter__(self):
        yield from self._annotations.values()

    def items(self):
        """A view on the image annotation items (key-value pairs)."""
        return self._annotations.items()

    def __contains__(self, annotation: Annotation) -> bool:
        """Return `True` if a given annotation is present in the dataset, else `False`."""
        return annotation.image_id in self._annotations.keys()

    def add(self, annotation: Annotation, *, override=False):
        """
        Add an annotation to the dataset.

        Parameters:

        * `annotation`: the annotation to add.
        * `override`: set to `True` if the annotation may already be in the dataset and the former
        it should be replaced by the new one. If `False` and the annotation is already in the
        dataset, an error is thrown.
        """

        if not override:
            assert annotation.image_id not in self.image_ids, (
                f"The annotation with id '{annotation.image_id}' is already present in the set "
                "(set `override` to True to remove this assertion)."
            )
        self._annotations[annotation.image_id] = annotation

    def update(self, other: "AnnotationSet", *, override=False) -> "AnnotationSet":
        """
        Add annotations from another datasetset to this one.

        Parameters:

        * `other`: the annotations to add.
        * `override`: if `True`, image annotations in `other` that aren't unique are allowed and
        only the last one in the iterator will be kept, else an error is thrown.
        """

        if not override:
            assert self.image_ids.isdisjoint(other.image_ids), (
                "some image ids are already in the set (set 'override' to True to remove "
                "this assertion)."
            )
        self._annotations.update(other._annotations)
        return self

    def __ior__(self, other: "AnnotationSet") -> "AnnotationSet":
        return self.update(other)

    def __or__(self, other: "AnnotationSet") -> "AnnotationSet":
        return AnnotationSet().update(self).update(other)

    def map_labels(self, mapping: Mapping[str, str]) -> "AnnotationSet":
        """
        Update all the bounding box labels according to the provided dictionary which maps former
        names to new names. If a label name is not present in the dictionary keys, then it won't
        be updated.
        """
        for annotation in self:
            annotation.map_labels(mapping)
        return self

    @property
    def image_ids(self):
        """A view on the set of image IDs of this dataset."""
        return self._annotations.keys()

    @property
    def all_boxes(self) -> Iterator[BoundingBox]:
        """An iterator of all the bounding boxes of the dataset."""
        for annotation in self:
            yield from annotation.boxes

    def nb_boxes(self) -> int:
        """The number of bounding boxes in the dataset."""
        return sum(len(ann.boxes) for ann in self)

    def _labels(self) -> "set[str]":
        """The set of the different label names present in the dataset."""
        return {b.label for b in self.all_boxes}

    @staticmethod
    def from_iter(
        parser: Callable[[T], Annotation],
        iterable: Iterable[T],
        *,
        verbose: bool = False,
    ) -> "AnnotationSet":
        annotations = thread_map(parser, iterable, desc="Parsing", verbose=verbose)
        return AnnotationSet(annotations)

    @staticmethod
    def from_folder(
        folder: PathLike,
        *,
        extension: str,
        parser: Callable[[Path], Annotation],
        recursive=False,
        verbose: bool = False,
    ) -> "AnnotationSet":
        folder = Path(folder).expanduser().resolve()

        assert (
            folder.is_dir()
        ), f"Filepath '{folder}' is not a folder or does not exist."

        files = list(glob(folder, extension, recursive=recursive))
        return AnnotationSet.from_iter(parser, files, verbose=verbose)

    @staticmethod
    def from_txt(
        folder: PathLike,
        *,
        image_folder: Optional[PathLike] = None,
        box_format=BoxFormat.LTRB,
        relative=False,
        file_extension: str = ".txt",
        image_extension: str = ".jpg",
        separator: Optional[str] = None,
        conf_last: bool = False,
        verbose: bool = False,
    ) -> "AnnotationSet":
        """This method won't try to retreive the image sizes by default. Specify `image_folder` if you need them.
        `image_folder` is required when `relative` is True."""
        # TODO: Add error handling

        folder = Path(folder).expanduser().resolve()

        assert folder.is_dir()
        assert image_extension.startswith(".")

        if relative:
            assert (
                image_folder is not None
            ), "When `relative` is set to True, `image_folder` must be provided to read image sizes."

        if image_folder is not None:
            image_folder = Path(image_folder).expanduser().resolve()
            assert image_folder.is_dir()

        def _get_annotation(file: Path) -> Annotation:
            if image_folder is not None:
                image_path: Path | None = None

                for image_ext in IMAGE_EXTENSIONS:
                    image_id = file.with_suffix(image_ext).name
                    path = image_folder / image_id  # type: ignore

                    if path.is_file():
                        image_path = path
                        break

                assert (
                    image_path is not None
                ), f"Image {file.name} does not exist, unable to read the image size."

                image_id = image_path.name

                try:
                    image_size = get_image_size(image_path)
                except UnknownImageFormat:
                    raise ParsingError(
                        f"Unable to read image size of file {image_path}. "
                        f"The file may be corrupted or the file format not supported."
                    )
            else:
                image_size = None
                image_id = file.with_suffix(image_extension).name

            return Annotation.from_txt(
                file_path=file,
                image_id=image_id,
                box_format=box_format,
                relative=relative,
                image_size=image_size,
                separator=separator,
                conf_last=conf_last,
            )

        return AnnotationSet.from_folder(
            folder,
            extension=file_extension,
            parser=_get_annotation,
            verbose=verbose,
        )

    @staticmethod
    def _from_yolo(
        folder: PathLike,
        *,
        image_folder: PathLike,
        image_extension=".jpg",
        conf_last: bool = False,
        verbose: bool = False,
    ) -> "AnnotationSet":
        return AnnotationSet.from_txt(
            folder,
            image_folder=image_folder,
            box_format=BoxFormat.XYWH,
            relative=True,
            image_extension=image_extension,
            separator=None,
            conf_last=conf_last,
            verbose=verbose,
        )

    @staticmethod
    def from_yolo(
        folder: PathLike,
        *,
        image_folder: PathLike,
        image_extension=".jpg",
        conf_last: bool = False,
        verbose: bool = False,
    ) -> "AnnotationSet":
        warn(
            "'from_yolo' is deprecated. Please use `from_yolo_darknet` or `from_yolo_v5`",
            category=DeprecationWarning,
            stacklevel=2,
        )

        return AnnotationSet._from_yolo(
            folder,
            image_folder=image_folder,
            image_extension=image_extension,
            conf_last=conf_last,
            verbose=verbose,
        )

    @staticmethod
    def from_yolo_darknet(
        folder: PathLike,
        *,
        image_folder: PathLike,
        image_extension=".jpg",
        verbose: bool = False,
    ) -> "AnnotationSet":
        return AnnotationSet._from_yolo(
            folder,
            image_folder=image_folder,
            image_extension=image_extension,
            conf_last=False,
            verbose=verbose,
        )

    @staticmethod
    def from_yolo_v5(
        folder: PathLike,
        *,
        image_folder: PathLike,
        image_extension=".jpg",
        verbose: bool = False,
    ) -> "AnnotationSet":
        return AnnotationSet._from_yolo(
            folder,
            image_folder=image_folder,
            image_extension=image_extension,
            conf_last=True,
            verbose=verbose,
        )

    @staticmethod
    def from_yolo_v7(
        folder: PathLike,
        *,
        image_folder: PathLike,
        image_extension=".jpg",
        verbose: bool = False,
    ) -> "AnnotationSet":
        return AnnotationSet.from_yolo_v5(
            folder,
            image_folder=image_folder,
            image_extension=image_extension,
            verbose=verbose,
        )

    @staticmethod
    def from_xml(folder: PathLike, *, verbose: bool = False) -> "AnnotationSet":
        return AnnotationSet.from_folder(
            folder, extension=".xml", parser=Annotation.from_xml, verbose=verbose
        )

    @staticmethod
    def from_pascal_voc(folder: PathLike, *, verbose: bool = False) -> "AnnotationSet":
        return AnnotationSet.from_xml(folder, verbose=verbose)

    @staticmethod
    def from_imagenet(folder: PathLike, *, verbose: bool = False) -> "AnnotationSet":
        return AnnotationSet.from_xml(folder, verbose=verbose)

    @staticmethod
    def from_openimage(
        file_path: PathLike,
        *,
        image_folder: PathLike,
        verbose: bool = False,
    ) -> "AnnotationSet":
        file_path = Path(file_path).expanduser().resolve()
        assert (
            file_path.is_file() and file_path.suffix == ".csv"
        ), f"OpenImage annotation file {file_path} must be a csv file."

        image_folder = Path(image_folder).expanduser().resolve()
        assert (
            image_folder.is_dir()
        ), f"Image folder {image_folder} must be a valid directory."

        # TODO: Error handling.
        # OSError, DictReader error, Key/Value Error

        annotations = AnnotationSet()

        with file_path.open(newline="") as f:
            reader = csv.DictReader(f)

            for row in tqdm(reader, desc="Parsing", disable=not verbose):
                image_id = row["ImageID"]
                label = row["LabelName"]
                coords = (float(row[r]) for r in ("XMin", "YMin", "XMax", "YMax"))
                confidence = row.get("Confidence")

                if confidence is not None and confidence != "":
                    confidence = float(confidence)
                else:
                    confidence = None

                if image_id not in annotations.image_ids:
                    image_path = image_folder / image_id
                    image_size = get_image_size(image_path)
                    annotations.add(
                        Annotation(image_id=image_id, image_size=image_size)
                    )

                annotation = annotations[image_id]
                annotation.add(
                    BoundingBox.create(
                        label=label,
                        coords=tuple(coords),
                        confidence=confidence,
                        relative=True,
                        image_size=annotation.image_size,
                    )
                )

        return annotations

    @staticmethod
    def from_labelme(
        folder: PathLike, *, include_poly: bool = False, verbose: bool = False
    ) -> "AnnotationSet":
        parser = partial(Annotation.from_labelme, include_poly=include_poly)
        return AnnotationSet.from_folder(
            folder, extension=".json", parser=parser, verbose=verbose
        )

    @staticmethod
    def from_coco(file_path: PathLike, *, verbose: bool = False) -> "AnnotationSet":
        file_path = Path(file_path).expanduser().resolve()
        assert (
            file_path.is_file() and file_path.suffix == ".json"
        ), f"COCO annotation file {file_path} must be a json file."

        # TODO: Error handling.
        # OSError, JsonDecoderError, Key/ValueError
        with file_path.open() as f:
            content = json.load(f)

        id_to_label = {d["id"]: str(d["name"]) for d in content["categories"]}

        id_to_annotation = {
            d["id"]: Annotation._from_coco_partial(d) for d in content["images"]
        }

        elements = content["annotations"]

        for element in tqdm(elements, desc="Parsing", disable=not verbose):
            annotation = id_to_annotation[element["image_id"]]
            label = id_to_label[int(element["category_id"])]
            coords = tuple(float(c) for c in element["bbox"])
            confidence = element.get("score")

            if confidence is not None:
                confidence = float(confidence)

            annotation.add(
                BoundingBox.create(
                    label=label,
                    coords=coords,
                    confidence=confidence,
                    box_format=BoxFormat.LTWH,
                )
            )

        annotation_set = AnnotationSet(id_to_annotation.values())
        annotation_set._id_to_label = id_to_label
        annotation_set._id_to_imageid = {
            idx: ann.image_id for idx, ann in id_to_annotation.items()
        }

        return annotation_set

    def from_results(
        self, file_path: PathLike, *, verbose: bool = False
    ) -> "AnnotationSet":
        file_path = Path(file_path).expanduser().resolve()
        # TODO: Error handling.
        assert (
            file_path.is_file() and file_path.suffix == ".json"
        ), f"COCO annotation file {file_path} must be a json file."

        id_to_label = self._id_to_label
        id_to_imageid = self._id_to_imageid

        assert id_to_label is not None and id_to_imageid is not None, (
            "The AnnotationSet instance should have been created with `AnnotationSet.from_coco()` "
            "or should have `self.id_to_label` and `self.id_to_image_id` populated. If not the "
            "case use the static method `AnnotationSet.from_coco_results()` instead."
        )

        id_to_annotation = {}

        with file_path.open() as f:
            annotations = json.load(f)

        # TODO: Factorize this with `Self.from_coco()`?
        for element in tqdm(annotations, desc="Parsing", disable=not verbose):
            image_id = id_to_imageid[element["image_id"]]
            gt_ann = self[image_id]

            if image_id not in id_to_annotation:
                annotation = Annotation(image_id, image_size=gt_ann.image_size)
                id_to_annotation[image_id] = annotation
            else:
                annotation = id_to_annotation[image_id]

            label = id_to_label[int(element["category_id"])]
            coords = tuple(float(c) for c in element["bbox"])
            confidence = float(element["score"])

            annotation.add(
                BoundingBox.create(
                    label=label,
                    coords=coords,
                    confidence=confidence,
                    box_format=BoxFormat.LTWH,
                )
            )

        annotation_set = AnnotationSet(id_to_annotation.values())
        annotation_set._id_to_label = id_to_label
        annotation_set._id_to_imageid = id_to_imageid

        return annotation_set

    @staticmethod
    def from_coco_results(
        file_path: PathLike,
        *,
        id_to_label: "dict[int, str]",
        id_to_imageid: "dict[int, str]",
        verbose: bool = False,
    ) -> "AnnotationSet":
        # TODO: Error handling.

        file_path = Path(file_path).expanduser().resolve()
        assert (
            file_path.is_file() and file_path.suffix == ".json"
        ), f"COCO annotation file {file_path} must be a json file."

        id_to_annotation = {}

        with file_path.open() as f:
            annotations = json.load(f)

        # TODO: Factorize this with `Self.from_coco()`?
        for element in tqdm(annotations, desc="Parsing", disable=not verbose):
            image_id = id_to_imageid[element["image_id"]]

            if image_id not in id_to_annotation:
                annotation = Annotation(image_id)
                id_to_annotation[image_id] = annotation
            else:
                annotation = id_to_annotation[image_id]

            label = id_to_label[int(element["category_id"])]
            coords = tuple(float(c) for c in element["bbox"])
            confidence = float(element["score"])

            annotation.add(
                BoundingBox.create(
                    label=label,
                    coords=coords,
                    confidence=confidence,
                    box_format=BoxFormat.LTWH,
                )
            )

        annotation_set = AnnotationSet(id_to_annotation.values())
        annotation_set._id_to_label = id_to_label
        annotation_set._id_to_imageid = id_to_imageid

        return annotation_set

    @staticmethod
    def from_cvat(file_path: PathLike, *, verbose: bool = False) -> "AnnotationSet":
        file_path = Path(file_path).expanduser().resolve()
        assert (
            file_path.is_file() and file_path.suffix == ".xml"
        ), f"CVAT annotation file {file_path} must be a xml file."

        # TODO: Error handling.
        with file_path.open() as f:
            root = et.parse(f).getroot()

        image_nodes = list(root.iter("image"))

        return AnnotationSet.from_iter(
            Annotation.from_cvat, image_nodes, verbose=verbose
        )

    @staticmethod
    def from_via_json(
        file_path: PathLike,
        *,
        label_key: str = "label_id",
        confidence_key: str = "confidence",
        image_folder: Optional[PathLike] = None,
    ) -> "AnnotationSet":
        file_path = Path(file_path).expanduser().resolve()

        if image_folder is not None:
            image_folder = Path(image_folder).expanduser().resolve()

            if not image_folder.is_dir():
                raise ParsingError("Invalid `image_folder`: not a directory.")

        if not file_path.is_file() or not file_path.suffix == ".json":
            raise ParsingError(
                f"VIA JSON annotation file {file_path} must be a valid json file."
            )

        with file_path.open() as f:
            content: dict = json.load(f)

        img_anns: dict = content.get("_via_img_metadata", content)

        annotations = AnnotationSet()

        for img_ann in img_anns.values():
            annotation = Annotation.from_via_json(
                img_ann, label_key=label_key, confidence_key=confidence_key
            )

            if image_folder is not None:
                img_path = image_folder / annotation.image_id
                image_size = get_image_size(img_path)
                annotation.image_size = image_size

            annotations.add(annotation)

        return annotations

    @staticmethod
    def from_yolo_seg(
        folder: PathLike,
        *,
        image_folder: Optional[PathLike] = None,
        relative=False,
        file_extension: str = ".txt",
        image_extension: str = ".jpg",
        verbose: bool = False,
    ) -> "AnnotationSet":
        """This method won't try to retreive the image sizes by default. Specify `image_folder` if you need them.
        `image_folder` is required when `relative` is True."""
        # TODO: Add error handling

        folder = Path(folder).expanduser().resolve()

        assert folder.is_dir()
        assert image_extension.startswith(".")

        if relative:
            assert (
                image_folder is not None
            ), "When `relative` is set to True, `image_folder` must be provided to read image sizes."

        if image_folder is not None:
            image_folder = Path(image_folder).expanduser().resolve()
            assert image_folder.is_dir()

        def _get_annotation(file: Path) -> Annotation:
            if image_folder is not None:
                image_path: Path | None = None

                for image_ext in IMAGE_EXTENSIONS:
                    image_id = file.with_suffix(image_ext).name
                    path = image_folder / image_id  # type: ignore

                    if path.is_file():
                        image_path = path
                        break

                assert (
                    image_path is not None
                ), f"Image {file.name} does not exist, unable to read the image size."

                image_id = image_path.name

                try:
                    image_size = get_image_size(image_path)
                except UnknownImageFormat:
                    raise ParsingError(
                        f"Unable to read image size of file {image_path}. "
                        f"The file may be corrupted or the file format not supported."
                    )
            else:
                image_size = None
                image_id = file.with_suffix(image_extension).name

            return Annotation.from_yolo_seg(
                file_path=file, image_id=image_id, image_size=image_size
            )

        return AnnotationSet.from_folder(
            folder,
            extension=file_extension,
            parser=_get_annotation,
            verbose=verbose,
        )

    def save_from_it(
        self, save_fn: Callable[[Annotation], None], *, verbose: bool = False
    ):
        thread_map(save_fn, self, desc="Saving", verbose=verbose)

    def save_txt(
        self,
        save_dir: PathLike,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        box_format: BoxFormat = BoxFormat.LTRB,
        relative: bool = False,
        separator: str = " ",
        file_extension: str = ".txt",
        conf_last: bool = False,
        verbose: bool = False,
    ):
        save_dir = Path(save_dir).expanduser().resolve()
        save_dir.mkdir(exist_ok=True)

        def _save(annotation: Annotation):
            image_id = annotation.image_id
            path = (save_dir / image_id).with_suffix(file_extension)

            annotation.save_txt(
                path,
                label_to_id=label_to_id,
                box_format=box_format,
                relative=relative,
                separator=separator,
                conf_last=conf_last,
            )

        self.save_from_it(_save, verbose=verbose)

    def _save_yolo(
        self,
        save_dir: PathLike,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        conf_last: bool = False,
        verbose: bool = False,
    ):
        save_dir = Path(save_dir).expanduser().resolve()
        save_dir.mkdir(exist_ok=True)

        def _save(annotation: Annotation):
            path = save_dir / Path(annotation.image_id).with_suffix(".txt")
            annotation._save_yolo(path, label_to_id=label_to_id, conf_last=conf_last)

        self.save_from_it(_save, verbose=verbose)

    def save_yolo(
        self,
        save_dir: PathLike,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        conf_last: bool = False,
        verbose: bool = False,
    ):
        warn(
            "'save_yolo' is deprecated. Please use `save_yolo_darknet` or `save_yolo_v5`",
            category=DeprecationWarning,
            stacklevel=2,
        )

        self._save_yolo(
            save_dir, label_to_id=label_to_id, conf_last=conf_last, verbose=verbose
        )

    def save_yolo_darknet(
        self,
        save_dir: PathLike,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        verbose: bool = False,
    ):
        self._save_yolo(
            save_dir, label_to_id=label_to_id, conf_last=False, verbose=verbose
        )

    def save_yolo_v5(
        self,
        save_dir: PathLike,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        verbose: bool = False,
    ):
        self._save_yolo(
            save_dir, label_to_id=label_to_id, conf_last=True, verbose=verbose
        )

    def save_yolo_v7(
        self,
        save_dir: PathLike,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        verbose: bool = False,
    ):
        self.save_yolo_v5(save_dir, label_to_id=label_to_id, verbose=verbose)

    def save_labelme(self, save_dir: PathLike, *, verbose: bool = False):
        save_dir = Path(save_dir).expanduser().resolve()
        save_dir.mkdir(exist_ok=True)

        def _save(annotation: Annotation):
            path = save_dir / Path(annotation.image_id).with_suffix(".json")
            annotation.save_labelme(path)

        self.save_from_it(_save, verbose=verbose)

    def save_xml(self, save_dir: PathLike, *, verbose: bool = False):
        save_dir = Path(save_dir).expanduser().resolve()
        save_dir.mkdir(exist_ok=True)

        def _save(annotation: Annotation):
            path = save_dir / Path(annotation.image_id).with_suffix(".xml")
            annotation.save_xml(path)

        self.save_from_it(_save, verbose=verbose)

    def save_pascal_voc(self, save_dir: PathLike, *, verbose: bool = False):
        self.save_xml(save_dir, verbose=verbose)

    def save_imagenet(self, save_dir: PathLike, *, verbose: bool = False):
        self.save_xml(save_dir, verbose=verbose)

    def to_coco(
        self,
        *,
        label_to_id: Optional["dict[str, int]"] = None,
        imageid_to_id: Optional["dict[str, int]"] = None,
        auto_ids: bool = False,
        verbose: bool = False,
    ) -> dict:
        if (label_to_id is not None) and (imageid_to_id is not None):
            pass
        elif (self._id_to_label is not None) and (self._id_to_imageid is not None):
            label_to_id = {v: k for k, v in self._id_to_label.items()}
            imageid_to_id = {v: k for k, v in self._id_to_imageid.items()}
        elif auto_ids:
            label_to_id = {l: i for i, l in enumerate(sorted(self._labels()))}
            imageid_to_id = {im: i for i, im in enumerate(sorted(self.image_ids))}
        else:
            # TODO: Convert to ConversionError.
            raise ValueError(
                "For COCO, mappings from labels and image ids to integer ids are required. "
                "They can be provided either by argument or automatically by the `AnnotationSet` "
                "instance if it was created with `AnnotationSet.from_coco()` or "
                "`AnnotationSet.from_coco_results()`. You can also set `auto_ids` to True to "
                "automatically create image and label ids (warning: this could cause unexpected "
                "compatibility issues with other COCO datasets)."
            )

        annotations = []
        ann_id_count = 0
        for annotation in tqdm(self, desc="Saving", disable=not verbose):
            for box in annotation.boxes:
                box_annotation = {
                    "iscrowd": 0,
                    "ignore": 0,
                    "image_id": imageid_to_id[annotation.image_id],
                    "bbox": box.ltwh,
                    "area": box.area,
                    "segmentation": [],
                    "category_id": label_to_id[box.label],
                    "id": ann_id_count,
                }

                if box.is_detection:
                    box_annotation["score"] = box.confidence

                annotations.append(box_annotation)

                ann_id_count += 1

        images = [
            {
                "id": imageid_to_id[a.image_id],
                "file_name": a.image_id,
                "width": a.image_width,
                "height": a.image_height,
            }
            for a in self
        ]

        categories = [
            {"supercategory": "none", "id": i, "name": l}
            for l, i in label_to_id.items()
        ]

        return {"images": images, "annotations": annotations, "categories": categories}

    def save_coco(
        self,
        path: PathLike,
        *,
        label_to_id: Optional["dict[str, int]"] = None,
        imageid_to_id: Optional["dict[str, int]"] = None,
        auto_ids: bool = False,
        verbose: bool = False,
    ):
        path = Path(path).expanduser().resolve()

        if path.suffix == "":
            path = path.with_suffix(".json")

        assert path.suffix == ".json", f"Path '{path}' suffix should be '.json'."

        content = self.to_coco(
            label_to_id=label_to_id,
            imageid_to_id=imageid_to_id,
            auto_ids=auto_ids,
            verbose=verbose,
        )

        with open_atomic(path, "w") as f:
            json.dump(content, fp=f, allow_nan=False)

    def save_openimage(self, path: PathLike, *, verbose: bool = False):
        path = Path(path).expanduser().resolve()

        if path.suffix == "":
            path = path.with_suffix(".csv")

        assert path.suffix == ".csv", f"Path '{path}' suffix should be '.csv'."

        fields = (
            "ImageID",
            "Source",
            "LabelName",
            "Confidence",
            "XMin",
            "XMax",
            "YMin",
            "YMax",
            "IsOccluded",
            "IsTruncated",
            "IsGroupOf",
            "IsDepiction",
            "IsInside",
        )

        with open_atomic(path, "w", newline="") as f:
            writer = csv.DictWriter(f, fieldnames=fields, restval="")
            writer.writeheader()

            for annotation in tqdm(self, desc="Saving", disable=not verbose):
                image_id = annotation.image_id
                image_size = annotation.image_size

                if image_size is None:
                    raise ValueError(
                        "The image size should be present in the annotation for `save_openimage`. "
                        "One should parse the annotations specifying the image folder or populate "
                        "the `image_size` attribute."
                    )

                for box in annotation.boxes:
                    label = box.label

                    if "," in label:
                        raise ValueError(
                            f"The box label '{label}' contains the character ',' which is the same "
                            "as the separtor character used for BoundingBox representation in "
                            "OpenImage format (CSV). This will corrupt the saved annotation file "
                            "and likely make it unreadable. Use another character in the label "
                            "name, e.g. use and underscore instead of a comma."
                        )

                    xmin, ymin, xmax, ymax = BoundingBox.abs_to_rel(
                        coords=box.ltrb, size=image_size
                    )

                    row = {
                        "ImageID": image_id,
                        "LabelName": label,
                        "XMin": xmin,
                        "XMax": xmax,
                        "YMin": ymin,
                        "YMax": ymax,
                    }

                    if box.confidence is not None:
                        row["Confidence"] = box.confidence

                    writer.writerow(row)  # type: ignore

    def to_cvat(self, *, verbose: bool = False) -> et.Element:
        def _create_node(annotation: Annotation) -> et.Element:
            return annotation.to_cvat()

        sub_nodes = thread_map(_create_node, self, desc="Saving", verbose=verbose)
        node = et.Element("annotations")
        node.extend(sub_nodes)

        return node

    def save_cvat(self, path: PathLike, *, verbose: bool = False):
        path = Path(path).expanduser().resolve()

        if path.suffix == "":
            path = path.with_suffix(".xml")

        assert path.suffix == ".xml", f"Path '{path}' suffix should be '.xml'."

        content = self.to_cvat(verbose=verbose)
        content = et.tostring(content, encoding="unicode")

        with open_atomic(path, "w") as f:
            f.write(content)

    def to_via_json(
        self,
        *,
        image_folder: Path,
        label_key: str = "label_id",
        confidence_key: str = "confidence",
        verbose: bool = False,
    ) -> dict:
        output = {}

        for annotation in tqdm(self, desc="Saving", disable=not verbose):
            ann_dict = annotation.to_via_json(
                image_folder=image_folder,
                label_key=label_key,
                confidence_key=confidence_key,
            )

            key = f"{ann_dict['filename']}{ann_dict['size']}"
            output[key] = ann_dict

        return output

    def save_via_json(
        self,
        path: PathLike,
        *,
        image_folder: Path,
        label_key: str = "label_id",
        confidence_key: str = "confidence",
        verbose: bool = False,
    ):
        path = Path(path).expanduser().resolve()

        if path.suffix == "":
            path = path.with_suffix(".json")

        assert path.suffix == ".json", f"Path '{path}' suffix should be '.json'."

        output = self.to_via_json(
            image_folder=image_folder,
            label_key=label_key,
            confidence_key=confidence_key,
            verbose=verbose,
        )

        with open_atomic(path, "w") as f:
            json.dump(output, fp=f)

    @staticmethod
    def parse_names_file(path: PathLike) -> "dict[str, str]":
        """Parse .names file.

        Parameters:
        - path: the path to the .names file.

        Returns:
        - A dictionary mapping label number with label names."""
        # TODO: Add error handling
        path = Path(path).expanduser().resolve()

        return {str(i): v for i, v in enumerate(path.read_text().splitlines())}

    @staticmethod
    def parse_mid_file(path: PathLike) -> "dict[str, str]":
        path = Path(path).expanduser().resolve()

        with path.open() as f:
            reader = csv.DictReader(f, fieldnames=("LabelName", "DisplayName"))
            return {l["LabelName"]: l["DisplayName"] for l in reader}

    def show_stats(self, *, verbose: bool = False):
        """
        Print in the console a synthetic view of the dataset annotations (distribution of
        bounding boxes and images by label).
        """
        from rich import print as rprint
        from rich.table import Table

        box_by_label = defaultdict(int)
        im_by_label = defaultdict(int)

        for annotation in tqdm(self, desc="Stats", disable=not verbose):
            for box in annotation.boxes:
                box_by_label[box.label] += 1

            labels = annotation._labels()

            for label in labels:
                im_by_label[label] += 1

            if len(labels) == 0:
                im_by_label["<empty_image>"] += 1

        tot_box = sum(box_by_label.values())
        tot_im = len(self)

        table = Table(title="Database Stats", show_footer=True)
        table.add_column("Label", footer="Total")
        table.add_column("Images", footer=f"{tot_im}", justify="right")
        table.add_column("Boxes", footer=f"{tot_box}", justify="right")

        for label in sorted(im_by_label.keys()):
            nb_im = im_by_label[label]
            nb_box = box_by_label[label]
            table.add_row(label, f"{nb_im}", f"{nb_box}")

        rprint(table)

    def __repr__(self) -> str:
        return f"AnnotationSet(annotations: {self._annotations})"



import json
import xml.etree.ElementTree as et
from pathlib import Path
from typing import Mapping, Optional, Union
from warnings import warn

from .atomic import open_atomic
from .boundingbox import BoundingBox, BoxFormat
from .errors import FileParsingError, ParsingError
from .file_utils import PathLike


class Annotation:
    """
    The bounding boxes associated with a uniquely identified image.
    """

    __slots__ = ("_image_id", "_image_size", "boxes")

    def __init__(
        self,
        image_id: str,
        image_size: Optional["tuple[int, int]"] = None,
        boxes: Optional["list[BoundingBox]"] = None,
    ) -> None:
        """
        Create an `Annotation` for an image identified with a unique `str` tag and with
        the provided list of bounding boxes.

        The image size in pixels ((width, height) tuple) can be optionally specified and is required
        for some export formats. This value can be queried from an image file with the
        `get_image_size()` function if it cannot be retreived from the annotation file.
        """
        if image_size is not None:
            img_w, img_h = image_size
            assert (
                img_w > 0 and img_h > 0
            ), f"Image size '({img_h}, {img_w})' should be positive."
            assert (
                int(img_w) == img_w and int(img_h) == img_h
            ), f"Image size '({img_h}, {img_w})' components should be integers."

        self._image_id = image_id
        self._image_size = image_size
        self.boxes = boxes or []

    @property
    def image_id(self) -> str:
        """The unique identifier of the image for this annotation."""
        return self._image_id

    @property
    def image_size(self) -> "Optional[tuple[int, int]]":
        """The image size in pixels ((width, height) tuple) if present."""
        return self._image_size

    @image_size.setter
    def image_size(self, image_size: "Optional[tuple[int, int]]"):
        if image_size is not None:
            img_w, img_h = image_size
            assert (
                img_w > 0 and img_h > 0
            ), f"Image size '({img_h}, {img_w})' should be positive."
            assert (
                int(img_w) == img_w and int(img_h) == img_h
            ), f"Image size '({img_h}, {img_w})' components should be integers."
        self._image_size = image_size

    @property
    def image_width(self) -> Optional[int]:
        """The image width in pixels."""
        return self.image_size[0] if self.image_size is not None else None

    @property
    def image_height(self) -> Optional[int]:
        """The image height in pixels."""
        return self.image_size[1] if self.image_size is not None else None

    def add(self, box: BoundingBox):
        """Add a bounding box to the image annotation."""
        self.boxes.append(box)

    def map_labels(self, mapping: Mapping[str, str]) -> "Annotation":
        """
        Update all the bounding box labels according to the provided dictionary which maps former
        names to new names. If a label name is not present in the dictionary keys, then it won't
        be updated.
        """
        for box in self.boxes:
            if box.label in mapping.keys():
                box.label = mapping[box.label]
        return self

    def _labels(self) -> "set[str]":
        """The set of the different label names present in the annotation."""
        return {b.label for b in self.boxes}

    @staticmethod
    def from_txt(
        file_path: PathLike,
        *,
        image_id: Optional[str] = None,
        image_extension: str = ".jpg",
        box_format: BoxFormat = BoxFormat.LTRB,
        relative: bool = False,
        image_size: Optional["tuple[int, int]"] = None,
        separator: Optional[str] = None,
        conf_last: bool = False,
    ) -> "Annotation":
        path = Path(file_path).expanduser().resolve()

        if image_id is None:
            assert image_extension.startswith(
                "."
            ), f"Image extension '{image_extension}' should start with a dot."
            image_id = path.with_suffix(image_extension).name

        try:
            lines = path.read_text().splitlines()
        except OSError:
            raise FileParsingError(path, reason="cannot read file")

        try:
            boxes = [
                BoundingBox.from_txt(
                    l,
                    box_format=box_format,
                    relative=relative,
                    image_size=image_size,
                    separator=separator,
                    conf_last=conf_last,
                )
                for l in lines
            ]
        except ParsingError as e:
            raise FileParsingError(path, e.reason)

        return Annotation(image_id, image_size, boxes)

    @staticmethod
    def _from_yolo(
        file_path: PathLike,
        *,
        image_size: "tuple[int, int]",
        image_id: Optional[str] = None,
        image_extension: str = ".jpg",
        conf_last: bool = False,
    ) -> "Annotation":
        return Annotation.from_txt(
            file_path,
            image_id=image_id,
            image_extension=image_extension,
            box_format=BoxFormat.XYWH,
            relative=True,
            image_size=image_size,
            separator=None,
            conf_last=conf_last,
        )

    @staticmethod
    def from_yolo(
        file_path: PathLike,
        *,
        image_size: "tuple[int, int]",
        image_id: Optional[str] = None,
        image_extension: str = ".jpg",
        conf_last: bool = False,
    ) -> "Annotation":
        warn(
            "'from_yolo' is deprecated. Please use `from_yolo_darknet` or `from_yolo_v5`",
            category=DeprecationWarning,
            stacklevel=2,
        )

        return Annotation._from_yolo(
            file_path,
            image_size=image_size,
            image_id=image_id,
            image_extension=image_extension,
            conf_last=conf_last,
        )

    @staticmethod
    def from_yolo_darknet(
        file_path: PathLike,
        *,
        image_size: "tuple[int, int]",
        image_id: Optional[str] = None,
        image_extension: str = ".jpg",
    ) -> "Annotation":
        return Annotation._from_yolo(
            file_path,
            image_size=image_size,
            image_id=image_id,
            image_extension=image_extension,
            conf_last=False,
        )

    @staticmethod
    def from_yolo_v5(
        file_path: PathLike,
        *,
        image_size: "tuple[int, int]",
        image_id: Optional[str] = None,
        image_extension: str = ".jpg",
    ) -> "Annotation":
        return Annotation._from_yolo(
            file_path,
            image_size=image_size,
            image_id=image_id,
            image_extension=image_extension,
            conf_last=True,
        )

    @staticmethod
    def from_yolo_v7(
        file_path: PathLike,
        *,
        image_size: "tuple[int, int]",
        image_id: Optional[str] = None,
        image_extension: str = ".jpg",
    ) -> "Annotation":
        return Annotation.from_yolo_v5(
            file_path,
            image_size=image_size,
            image_id=image_id,
            image_extension=image_extension,
        )

    @staticmethod
    def from_xml(file_path: PathLike) -> "Annotation":
        path = Path(file_path).expanduser().resolve()

        try:
            with path.open() as f:
                root = et.parse(f).getroot()
        except (OSError, et.ParseError):
            raise ParsingError("Syntax error in imagenet annotation file.")

        image_id = root.findtext("filename")
        size_node = root.find("size")

        if (image_id is None) or (size_node is None):
            raise ParsingError("Syntax error in imagenet annotation file.")

        width = size_node.findtext("width")
        height = size_node.findtext("height")

        if (width is None) or (height is None):
            raise ParsingError("Syntax error in imagenet annotation file.")

        try:
            image_size = int(width), int(height)
        except ValueError:
            raise ParsingError("Syntax error in imagenet annotation file.")

        boxes = [BoundingBox.from_xml(n) for n in root.iter("object")]

        return Annotation(image_id, image_size, boxes)

    @staticmethod
    def from_pascal_voc(file_path: PathLike) -> "Annotation":
        return Annotation.from_xml(file_path)

    @staticmethod
    def from_imagenet(file_path: PathLike) -> "Annotation":
        return Annotation.from_xml(file_path)

    @staticmethod
    def from_labelme(file_path: PathLike, include_poly: bool = False) -> "Annotation":
        path = Path(file_path).expanduser().resolve()

        shape_types = {"rectangle"}
        if include_poly:
            shape_types.add("polygon")

        try:
            with path.open() as f:
                content = json.load(f)
                if "imageData" in content:
                    del content["imageData"]
        except (OSError, json.JSONDecodeError):
            raise ParsingError("Syntax error in labelme annotation file.")

        try:
            image_id = str(content["imagePath"])
            width = int(content["imageWidth"])
            height = int(content["imageHeight"])
            boxes = [
                BoundingBox.from_labelme(n)
                for n in content["shapes"]
                if n["shape_type"] in shape_types
            ]
        except (KeyError, ValueError):
            raise ParsingError("Syntax error in labelme annotation file.")

        return Annotation(image_id, image_size=(width, height), boxes=boxes)

    @staticmethod
    def _from_coco_partial(node: dict) -> "Annotation":
        try:
            image_id = str(node["file_name"])
            image_size = int(node["width"]), int(node["height"])
        except (ValueError, KeyError):
            raise ParsingError("Syntax error in COCO annotation file.")

        return Annotation(image_id, image_size)

    @staticmethod
    def from_cvat(node: et.Element) -> "Annotation":
        image_id = node.get("name")
        width, height = node.get("width"), node.get("height")

        if (image_id is None) or (width is None) or (height is None):
            raise ParsingError("Syntax error in CVAT annotation file.")

        try:
            img_size = int(width), int(height)
        except ValueError:
            raise ParsingError("Syntax error in CVAT annotation file.")

        boxes = [BoundingBox.from_cvat(n) for n in node.iter("box")]

        return Annotation(image_id, image_size=img_size, boxes=boxes)

    @staticmethod
    def from_via_json(
        annotation: dict,
        *,
        label_key: str = "label_id",
        confidence_key: str = "confidence",
        image_size: "Optional[tuple[int, int]]" = None,
    ) -> "Annotation":
        try:
            filename = annotation["filename"]
            regions = annotation["regions"]

            bboxes = [
                BoundingBox.from_via_json(
                    region, label_key=label_key, confidence_key=confidence_key
                )
                for region in regions
                if region["shape_attributes"]["name"] == "rect"
            ]
        except KeyError:
            raise ParsingError("Syntax error in VIA JSON annotation file.")

        return Annotation(
            image_id=filename,
            image_size=image_size,
            boxes=bboxes,
        )

    @staticmethod
    def from_yolo_seg(
        file_path: PathLike,
        *,
        image_id: Optional[str] = None,
        image_extension: str = ".jpg",
        image_size: Optional["tuple[int, int]"] = None,
    ) -> "Annotation":
        path = Path(file_path).expanduser().resolve()

        if image_id is None:
            assert image_extension.startswith(
                "."
            ), f"Image extension '{image_extension}' should start with a dot."
            image_id = path.with_suffix(image_extension).name

        try:
            lines = path.read_text().splitlines()
        except OSError:
            raise FileParsingError(path, reason="cannot read file")

        try:
            boxes = [
                BoundingBox.from_yolo_seg(
                    l,
                    image_size=image_size,
                )
                for l in lines
            ]
        except ParsingError as e:
            raise FileParsingError(path, e.reason)

        return Annotation(image_id, image_size, boxes)

    def to_txt(
        self,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        box_format: BoxFormat = BoxFormat.LTRB,
        relative=False,
        image_size: Optional["tuple[int, int]"] = None,
        separator: str = " ",
        conf_last: bool = False,
    ) -> str:
        image_size = image_size or self.image_size

        return "\n".join(
            box.to_txt(
                label_to_id=label_to_id,
                box_format=box_format,
                relative=relative,
                image_size=image_size,
                separator=separator,
                conf_last=conf_last,
            )
            for box in self.boxes
        )

    def _to_yolo(
        self,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        image_size: Optional["tuple[int, int]"] = None,
        conf_last: bool = False,
    ) -> str:
        image_size = image_size or self.image_size

        if image_size is None:
            raise ValueError(
                "Either `image_size` shoud be provided as argument or stored in the Annotation "
                "object for conversion to YOLO format."
            )

        return "\n".join(
            box.to_yolo(
                image_size=image_size, label_to_id=label_to_id, conf_last=conf_last
            )
            for box in self.boxes
        )

    def to_yolo(
        self,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        image_size: Optional["tuple[int, int]"] = None,
        conf_last: bool = False,
    ) -> str:
        warn(
            "'to_yolo' is deprecated. Please use `to_yolo_darknet` or `to_yolo_v5`",
            category=DeprecationWarning,
            stacklevel=2,
        )

        return self._to_yolo(
            label_to_id=label_to_id, image_size=image_size, conf_last=conf_last
        )

    def to_yolo_darknet(
        self,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        image_size: Optional["tuple[int, int]"] = None,
    ) -> str:
        return self._to_yolo(
            label_to_id=label_to_id, image_size=image_size, conf_last=False
        )

    def to_yolo_v5(
        self,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        image_size: Optional["tuple[int, int]"] = None,
    ) -> str:
        return self._to_yolo(
            label_to_id=label_to_id, image_size=image_size, conf_last=True
        )

    def to_yolo_v7(
        self,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        image_size: Optional["tuple[int, int]"] = None,
    ) -> str:
        return self.to_yolo_v5(label_to_id=label_to_id, image_size=image_size)

    def save_txt(
        self,
        path: PathLike,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        box_format: BoxFormat = BoxFormat.LTRB,
        relative: bool = False,
        image_size: Optional["tuple[int, int]"] = None,
        separator: str = " ",
        conf_last: bool = False,
    ):
        content = self.to_txt(
            label_to_id=label_to_id,
            box_format=box_format,
            relative=relative,
            image_size=image_size,
            separator=separator,
            conf_last=conf_last,
        )

        with open_atomic(path, "w") as f:
            f.write(content)

    def _save_yolo(
        self,
        path: PathLike,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        image_size: Optional["tuple[int, int]"] = None,
        conf_last: bool = False,
    ):
        content = self._to_yolo(
            label_to_id=label_to_id, image_size=image_size, conf_last=conf_last
        )

        with open_atomic(path, "w") as f:
            f.write(content)

    def save_yolo(
        self,
        path: PathLike,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        image_size: Optional["tuple[int, int]"] = None,
        conf_last: bool = False,
    ):
        warn(
            "'save_yolo' is deprecated. Please use `save_yolo_darknet` or `save_yolo_v5`",
            category=DeprecationWarning,
            stacklevel=2,
        )

        self._save_yolo(
            path, label_to_id=label_to_id, image_size=image_size, conf_last=conf_last
        )

    def save_yolo_darknet(
        self,
        path: PathLike,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        image_size: Optional["tuple[int, int]"] = None,
    ):
        self._save_yolo(
            path, label_to_id=label_to_id, image_size=image_size, conf_last=False
        )

    def save_yolo_v5(
        self,
        path: PathLike,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        image_size: Optional["tuple[int, int]"] = None,
    ):
        self._save_yolo(
            path, label_to_id=label_to_id, image_size=image_size, conf_last=True
        )

    def save_yolo_v7(
        self,
        path: PathLike,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        image_size: Optional["tuple[int, int]"] = None,
    ):
        self.save_yolo_v5(path, label_to_id=label_to_id, image_size=image_size)

    def to_labelme(self, *, image_size: Optional["tuple[int, int]"] = None) -> dict:
        image_size = image_size or self.image_size
        assert (
            image_size is not None
        ), "An image size should be provided either by argument or by `self.image_size`."

        return {
            "imagePath": self.image_id,
            "imageWidth": image_size[0],
            "imageHeight": image_size[1],
            "imageData": None,
            "shapes": [b.to_labelme() for b in self.boxes],
        }

    def save_labelme(
        self, path: PathLike, *, image_size: Optional["tuple[int, int]"] = None
    ):
        content = self.to_labelme(image_size=image_size)
        with open_atomic(path, "w") as f:
            json.dump(content, fp=f, allow_nan=False)

    def to_xml(self, *, image_size: Optional["tuple[int, int]"] = None) -> et.Element:
        image_size = image_size or self.image_size
        assert (
            image_size is not None
        ), "An image size should be provided either by argument or by `self.image_size`."

        ann_node = et.Element("annotation")
        et.SubElement(ann_node, "filename").text = self.image_id

        size_node = et.SubElement(ann_node, "size")
        et.SubElement(size_node, "width").text = f"{image_size[0]}"
        et.SubElement(size_node, "height").text = f"{image_size[1]}"

        for box in self.boxes:
            ann_node.append(box.to_xml())

        return ann_node

    def to_pascal_voc(
        self, *, image_size: Optional["tuple[int, int]"] = None
    ) -> et.Element:
        return self.to_xml(image_size=image_size)

    def to_imagenet(
        self, *, image_size: Optional["tuple[int, int]"] = None
    ) -> et.Element:
        return self.to_xml(image_size=image_size)

    def save_xml(
        self, path: PathLike, *, image_size: Optional["tuple[int, int]"] = None
    ):
        content = self.to_xml(image_size=image_size)
        content = et.tostring(content, encoding="unicode")

        with open_atomic(path, "w") as f:
            f.write(content)

    def save_pascal_voc(
        self, path: PathLike, *, image_size: Optional["tuple[int, int]"] = None
    ):
        self.save_xml(path, image_size=image_size)

    def save_imagenet(
        self, path: PathLike, *, image_size: Optional["tuple[int, int]"] = None
    ):
        self.save_xml(path, image_size=image_size)

    def to_cvat(self, *, image_size: Optional["tuple[int, int]"] = None) -> et.Element:
        image_size = image_size or self.image_size
        assert (
            image_size is not None
        ), "An image size should be provided either by argument or by `self.image_size`."

        img_node = et.Element(
            "image",
            attrib={
                "name": self.image_id,
                "width": f"{image_size[0]}",
                "height": f"{image_size[1]}",
            },
        )

        img_node.extend(box.to_cvat() for box in self.boxes)

        return img_node

    def to_via_json(
        self,
        *,
        image_folder: PathLike,
        label_key: str = "label_id",
        confidence_key: str = "confidence",
    ) -> dict:
        path = Path(image_folder).expanduser().resolve()

        assert path.is_dir(), f"Filepath '{path}' is not a folder or does not exist."

        image_id = self.image_id
        image_path = path / image_id
        file_size = image_path.stat().st_size

        regions = [
            box.to_via_json(label_key=label_key, confidence_key=confidence_key)
            for box in self.boxes
        ]

        return {"filename": image_id, "size": file_size, "regions": regions}

    def __repr__(self) -> str:
        return (
            f"Annotation(image_id: {self.image_id}, image_size: {self.image_size}, "
            f"boxes: {self.boxes})"
        )



from pathlib import Path
from typing import Iterable, Union

PathLike = Union[str, Path]


def glob(
    folder: PathLike, extensions: Union[str, Iterable[str]], recursive: bool = False
) -> Iterable[Path]:
    """Glob files by providing extensions to match."""
    if isinstance(extensions, str):
        extensions = {extensions}
    else:
        extensions = set(extensions)

    assert all(
        e.startswith(".") for e in extensions
    ), f"Parameter `extensions' ({extensions}) should all start with a dot."

    path = Path(folder).expanduser().resolve()

    files = path.glob("**/*") if recursive else path.glob("*")

    return (f for f in files if f.suffix in extensions and not f.name.startswith("."))



# Source: https://github.com/scardine/image_size

from os import path
from pathlib import Path
from struct import error as struct_error
from struct import unpack

from .errors import UnknownImageFormat
from .file_utils import PathLike

IMAGE_EXTENSIONS = [
    ".jpg",
    ".jpeg",
    ".png",
    ".bmp",
    ".jpe",
    ".tif",
    ".tiff",
    ".JPG",
    ".JPEG",
    ".PNG",
    ".BMP",
    ".JPE",
    ".TIF",
    ".TIFF",
]


def get_image_size(file_path: PathLike) -> "tuple[int, int]":
    """
    Compute the size of an image without loading into memory, which could result in faster speed.

    Parameters:

    * `file_path`: path to an image file.

    Returns:

    * The image size (width, height).
    """
    file_path = Path(file_path).expanduser().resolve()

    size = path.getsize(file_path)

    # be explicit with open arguments - we need binary mode
    with file_path.open("rb") as input:
        try:
            return _get_image_metadata_from_bytesio(input, size)
        except Exception as e:
            raise UnknownImageFormat(str(e))


def _get_image_metadata_from_bytesio(input, size: int) -> "tuple[int, int]":
    """
    Args:
        input (io.IOBase): io object support read & seek
        size (int): size of buffer in byte
        file_path (str): path to an image file
    Returns:
        Image: (path, type, file_size, width, height)
    """
    height = -1
    width = -1
    data = input.read(26)
    msg = " raised while trying to decode as JPEG."

    if (size >= 10) and data[:6] in (b"GIF87a", b"GIF89a"):
        # GIFs
        w, h = unpack("<HH", data[6:10])
        width = int(w)
        height = int(h)
    elif (
        (size >= 24)
        and data.startswith(b"\211PNG\r\n\032\n")
        and (data[12:16] == b"IHDR")
    ):
        # PNGs
        w, h = unpack(">LL", data[16:24])
        width = int(w)
        height = int(h)
    elif (size >= 16) and data.startswith(b"\211PNG\r\n\032\n"):
        # older PNGs
        w, h = unpack(">LL", data[8:16])
        width = int(w)
        height = int(h)
    elif (size >= 2) and data.startswith(b"\377\330"):
        # JPEG
        input.seek(0)
        input.read(2)
        b = input.read(1)
        try:
            while b and ord(b) != 0xDA:
                while ord(b) != 0xFF:
                    b = input.read(1)
                while ord(b) == 0xFF:
                    b = input.read(1)
                if ord(b) >= 0xC0 and ord(b) <= 0xC3:
                    input.read(3)
                    h, w = unpack(">HH", input.read(4))
                    width = int(w)
                    height = int(h)
                    break
                else:
                    input.read(int(unpack(">H", input.read(2))[0]) - 2)
                b = input.read(1)
        except struct_error:
            raise UnknownImageFormat("StructError" + msg)
        except ValueError:
            raise UnknownImageFormat("ValueError" + msg)
        except Exception as e:
            raise UnknownImageFormat(e.__class__.__name__ + msg)
    elif (size >= 26) and data.startswith(b"BM"):
        # BMP
        headersize = unpack("<I", data[14:18])[0]
        if headersize == 12:
            w, h = unpack("<HH", data[18:22])
            width = int(w)
            height = int(h)
        elif headersize >= 40:
            w, h = unpack("<ii", data[18:26])
            width = int(w)
            # as h is negative when stored upside down
            height = abs(int(h))
        else:
            raise UnknownImageFormat("Unkown DIB header size:" + str(headersize))
    elif (size >= 8) and data[:4] in (b"II\052\000", b"MM\000\052"):
        # Standard TIFF, big- or little-endian
        # BigTIFF and other different but TIFF-like formats are not
        # supported currently
        byteOrder = data[:2]
        boChar = ">" if byteOrder == "MM" else "<"
        # maps TIFF type id to size (in bytes)
        # and python format char for struct
        tiffTypes = {
            1: (1, boChar + "B"),  # BYTE
            2: (1, boChar + "c"),  # ASCII
            3: (2, boChar + "H"),  # SHORT
            4: (4, boChar + "L"),  # LONG
            5: (8, boChar + "LL"),  # RATIONAL
            6: (1, boChar + "b"),  # SBYTE
            7: (1, boChar + "c"),  # UNDEFINED
            8: (2, boChar + "h"),  # SSHORT
            9: (4, boChar + "l"),  # SLONG
            10: (8, boChar + "ll"),  # SRATIONAL
            11: (4, boChar + "f"),  # FLOAT
            12: (8, boChar + "d"),  # DOUBLE
        }
        ifdOffset = unpack(boChar + "L", data[4:8])[0]
        try:
            countSize = 2
            input.seek(ifdOffset)
            ec = input.read(countSize)
            ifdEntryCount = unpack(boChar + "H", ec)[0]
            # 2 bytes: TagId + 2 bytes: type + 4 bytes: count of values + 4
            # bytes: value offset
            ifdEntrySize = 12
            for i in range(ifdEntryCount):
                entryOffset = ifdOffset + countSize + i * ifdEntrySize
                input.seek(entryOffset)
                tag = input.read(2)
                tag = unpack(boChar + "H", tag)[0]
                if tag == 256 or tag == 257:
                    # if type indicates that value fits into 4 bytes, value
                    # offset is not an offset but value itself
                    type = input.read(2)
                    type = unpack(boChar + "H", type)[0]
                    if type not in tiffTypes:
                        raise UnknownImageFormat("Unkown TIFF field type:" + str(type))
                    typeSize = tiffTypes[type][0]
                    typeChar = tiffTypes[type][1]
                    input.seek(entryOffset + 8)
                    value = input.read(typeSize)
                    value = int(unpack(typeChar, value)[0])
                    if tag == 256:
                        width = value
                    else:
                        height = value
                if width > -1 and height > -1:
                    break
        except Exception as e:
            raise UnknownImageFormat(str(e))
    elif size >= 2:
        # see http://en.wikipedia.org/wiki/ICO_(file_format)
        input.seek(0)
        reserved = input.read(2)
        if 0 != unpack("<H", reserved)[0]:
            raise UnknownImageFormat("Sorry, don't know how to get size for this file")
        format = input.read(2)
        assert 1 == unpack("<H", format)[0]
        num = input.read(2)
        num = unpack("<H", num)[0]
        if num > 1:
            import warnings

            warnings.warn("ICO File contains more than one image")
        # http://msdn.microsoft.com/en-us/library/ms997538.aspx
        w = input.read(1)
        h = input.read(1)
        width = ord(w)
        height = ord(h)
    else:
        raise UnknownImageFormat("Sorry, don't know how to get size for this file")

    return width, height



from pathlib import Path


class UnknownImageFormat(Exception):
    pass


class ParsingError(Exception):
    def __init__(self, reason: str) -> None:
        self.reason = reason

    def __str__(self) -> str:
        return self.reason


class FileParsingError(ParsingError):
    def __init__(self, file: Path, reason: str) -> None:
        self.file = file
        self.reason = reason

    def __str__(self) -> str:
        return f"Error while reading file '{self.file}': {self.reason}"



from .annotation import Annotation
from .annotationset import AnnotationSet
from .boundingbox import BoundingBox, BoxFormat, Coordinates
from .errors import FileParsingError, ParsingError, UnknownImageFormat
from .evaluation import COCOEvaluator, Evaluation, EvaluationItem

__all__ = [
    "Annotation",
    "AnnotationSet",
    "BoundingBox",
    "BoxFormat",
    "Coordinates",
    "FileParsingError",
    "ParsingError",
    "UnknownImageFormat",
    "COCOEvaluator",
    "Evaluation",
    "EvaluationItem",
]



import os
import tempfile as tmp
from contextlib import contextmanager
from typing import Optional

from .file_utils import PathLike


@contextmanager
def _tempfile(suffix: str = "~", dir: Optional[PathLike] = None):
    tmp_file = tmp.NamedTemporaryFile(delete=False, suffix=suffix, dir=dir)
    tmp_name = tmp_file.name
    tmp_file.file.close()

    try:
        yield tmp_name
    finally:
        try:
            os.remove(tmp_name)
        except OSError as e:
            if e.errno == 2:
                pass
            else:
                raise


@contextmanager
def open_atomic(file_path: PathLike, *args, **kwargs):
    fsync = kwargs.pop("fsync", False)

    with _tempfile(dir=os.path.dirname(os.path.abspath(file_path))) as tmp_path:
        with open(tmp_path, *args, **kwargs) as file:
            try:
                yield file
            finally:
                if fsync:
                    file.flush()
                    os.fsync(file.fileno())

        os.replace(tmp_path, file_path)



import xml.etree.ElementTree as et
from enum import Enum, auto
from typing import Any, Mapping, Optional, Tuple, Union

from .errors import ParsingError

Coordinates = Tuple[float, float, float, float]
"""
The four raw coordinates of a `BoundingBox` with no assigned format. The coordinates at indices
0 and 2 must be on the X-axis (horizontal) and the ones at indices 1 and 3 must be on 
the Y-axis (vertical).
"""


class BoxFormat(Enum):
    """
    The coordinate format of a `BoundingBox`, i.e. the order and defining coordinates.

    It can be one of:

    * `BoxFormat.LTRB`: [xmin, ymin, xmax, ymax],
    * `BoxFormat.LTWH`: [xmin, ymin, width, height],
    * `BoxFormat.XYWH`: [xmid, ymid, width, height].

    See the `BoundingBox` documentation for more detail on the coordinate format.
    """

    LTRB = auto()
    """[xmin, ymin, xmax, ymax]"""

    LTWH = auto()
    """[xmin, ymin, width, height]"""

    XYWH = auto()
    """[xmid, ymid, width, height]"""

    @classmethod
    def from_string(cls, string: str) -> "BoxFormat":
        """
        Create a `BoxFormat` from a raw string.

        The raw string can be either "ltrb", "ltwh" or "xywh".
        """
        if string == "ltrb":
            return BoxFormat.LTRB
        elif string == "ltwh":
            return BoxFormat.LTWH
        elif string == "xywh":
            return BoxFormat.XYWH
        else:
            raise ValueError(f"Invalid BoxFormat string '{string}'")


class BoundingBox:
    """
    A rectangular bounding box with a label and an optional confidence score. Coordinates are
    absolute (i.e. in pixels) and stored internally in `BoxFormat.LTRB` format.

    Spatial layout:

    ```
        xmin   xmid   xmax
    ymin ╆╍╍╍╍╍╍┿╍╍╍╍╍╍┪
         ╏      ┆      ╏
    ymid ╂┄┄┄┄┄┄┼┄┄┄┄┄┄┨
         ╏      ┆      ╏
    ymax ┺╍╍╍╍╍╍┴╍╍╍╍╍╍┛
    ```
    """

    __slots__ = ("label", "_xmin", "_ymin", "_xmax", "_ymax", "_confidence")

    def __init__(
        self,
        *,
        label: str,
        xmin: float,
        ymin: float,
        xmax: float,
        ymax: float,
        confidence: Optional[float] = None,
    ) -> None:
        """
        Create a `BoundingBox` from the top-left and bottom-right corner coordinates.

        Use `BoundingBox.create()` to instantiate a `BoundingBox` from other
        coordinate formats.
        """
        assert xmin <= xmax, "`xmax` must be greater than `xmin`."
        assert ymin <= ymax, "`ymax` must be greater than `ymin`."

        if confidence is not None:
            assert (
                0.0 <= confidence <= 1.0
            ), f"Confidence ({confidence}) should be in [0, 1]."

        self.label = label
        self._xmin = xmin
        self._ymin = ymin
        self._xmax = xmax
        self._ymax = ymax
        self._confidence = confidence

    @property
    def confidence(self) -> Optional[float]:
        """
        The bounding box optional confidence score. If present, it is a `float`
        between 0 and 1.
        """
        return self._confidence

    @confidence.setter
    def confidence(self, confidence: Optional[float]):
        if confidence is not None:
            assert (
                0.0 <= confidence <= 1.0
            ), f"Confidence ({confidence}) should be in [0, 1]."
        self._confidence = confidence

    @property
    def xmin(self) -> float:
        """The `x` value in pixels of the bounding box top-left corner (horizontal axis)."""
        return self._xmin

    @property
    def ymin(self) -> float:
        """The `y` value in pixels of the bounding box top-left corner (vertical axis)."""
        return self._ymin

    @property
    def xmax(self) -> float:
        """The `x` value in pixels of the bounding box bottom-right corner (horizontal axis)."""
        return self._xmax

    @property
    def ymax(self) -> float:
        """The `y` value in pixels of the bounding box top-left corner (vertical axis)."""
        return self._ymax

    @property
    def xmid(self) -> float:
        """The `x` value in pixels of the bounding box center point (horizontal axis)."""
        return (self._xmin + self._xmax) / 2.0

    @property
    def ymid(self) -> float:
        """The `y` value in pixels of the bounding box center point (vertical axis)."""
        return (self._ymin + self._ymax) / 2.0

    @property
    def width(self) -> float:
        """The bounding box width in pixels."""
        return self._xmax - self._xmin

    @property
    def height(self) -> float:
        """The bounding box height in pixels."""
        return self._ymax - self._ymin

    @property
    def area(self) -> float:
        """The area of the bounding box in pixels."""
        return self.width * self.height

    def _area_in(self, range_: "tuple[float, float]") -> bool:
        """Returns `True` if the bounding box area is in a given range."""
        lower_bound, upper_bound = range_
        return lower_bound <= self.area <= upper_bound

    @property
    def pascal_area(self) -> int:
        """The bounding box PascalVOC area in pixels."""
        width = int(self._xmax) - int(self._xmin) + 1
        height = int(self._ymax) - int(self._ymin) + 1

        return width * height

    def iou(self, other: "BoundingBox") -> float:
        """The Intersection over Union (IoU) between two bounding boxes."""
        xmin = max(self._xmin, other._xmin)
        ymin = max(self._ymin, other._ymin)
        xmax = min(self._xmax, other._xmax)
        ymax = min(self._ymax, other._ymax)

        if xmax < xmin or ymax < ymin:
            return 0.0

        intersection = (xmax - xmin) * (ymax - ymin)
        union = self.area + other.area - intersection

        if union == 0.0:
            return 1.0

        return intersection / union

    def pascal_iou(self, other: "BoundingBox") -> float:
        """The Pascal VOC Intersection over Union (IoU) between two bounding boxes."""
        xmin = max(int(self._xmin), int(other._xmin))
        ymin = max(int(self._ymin), int(other._ymin))
        xmax = min(int(self._xmax), int(other._xmax))
        ymax = min(int(self._ymax), int(other._ymax))

        if xmax < xmin or ymax < ymin:
            return 0.0

        intersection = (xmax - xmin + 1) * (ymax - ymin + 1)
        union = self.pascal_area + other.pascal_area - intersection

        if union == 0:
            return 1.0

        return intersection / union

    @property
    def is_detection(self) -> bool:
        """Return `True` if the bounding box confidence score is not `None`."""
        return self._confidence is not None

    @property
    def is_ground_truth(self) -> bool:
        """Return `True` if the bounding box confidence score is `None`."""
        return self._confidence is None

    @staticmethod
    def rel_to_abs(coords: Coordinates, size: "tuple[int, int]") -> Coordinates:
        """
        Convert coordinates from relative (between 0 and 1) to absolute (pixels) form.

        The coordinates at indices 0 and 2 must be on the X-axis (horizontal) and the ones
        at indices 1 and 3 must be on the Y-axis (vertical).

        The image size should be a `(width, height)` tuple expressed in pixels.
        """
        a, b, c, d = coords
        w, h = size
        return a * w, b * h, c * w, d * h

    @staticmethod
    def abs_to_rel(coords: Coordinates, size: "tuple[int, int]") -> Coordinates:
        """
        Convert coordinates from absolute (pixels) to relative (between 0 and 1) form.

        The coordinates at indices 0 and 2 must be on the X-axis (horizontal) and the ones
        at indices 1 and 3 must be on the Y-axis (vertical).

        The image size should be a `(width, height)` tuple expressed in pixels.
        """
        a, b, c, d = coords
        w, h = size
        return a / w, b / h, c / w, d / h

    @staticmethod
    def ltwh_to_ltrb(coords: Coordinates) -> Coordinates:
        """
        Convert coordinates from `BoxFormat.LTWH` to `BoxFormat.LTRB` format.

        The coordinates can be either in relative (between 0 and 1) or absolute (pixels) form.
        """
        xmin, ymin, width, height = coords
        return xmin, ymin, xmin + width, ymin + height

    @staticmethod
    def xywh_to_ltrb(coords: Coordinates) -> Coordinates:
        """
        Convert coordinates from `BoxFormat.XYWH` to `BoxFormat.LTRB` format.

        The coordinates can be either in relative (between 0 and 1)  or absolute (pixels) form.
        """
        xmid, ymid, width, height = coords
        w_h, h_h = width / 2, height / 2
        return xmid - w_h, ymid - h_h, xmid + w_h, ymid + h_h

    @property
    def ltrb(self) -> Coordinates:
        """The bounding box coordinates in `BoxFormat.LTRB` format and absolute form (pixels)."""
        return self._xmin, self._ymin, self._xmax, self._ymax

    @property
    def ltwh(self) -> Coordinates:
        """The bounding box coordinates in `BoxFormat.LTWH` format and absolute form (pixels)."""
        return self._xmin, self._ymin, self.width, self.height

    @property
    def xywh(self) -> Coordinates:
        """The bounding box coordinates in `BoxFormat.XYWH` format and absolute form (pixels)."""
        return self.xmid, self.ymid, self.width, self.height

    @classmethod
    def create(
        cls,
        *,
        label: str,
        coords: Coordinates,
        confidence: Optional[float] = None,
        box_format=BoxFormat.LTRB,
        relative=False,
        image_size: Optional["tuple[int, int]"] = None,
    ) -> "BoundingBox":
        """
        Create a `BoundingBox` from different coordinate formats.

        The image size should be provided if the coordinates are given in the relative form
        (values between 0 and 1).
        """
        if relative:
            assert (
                image_size is not None
            ), "For relative coordinates `image_size` should be provided."
            coords = cls.rel_to_abs(coords, image_size)

        if box_format is BoxFormat.LTWH:
            coords = cls.ltwh_to_ltrb(coords)
        elif box_format is BoxFormat.XYWH:
            coords = cls.xywh_to_ltrb(coords)
        elif box_format is BoxFormat.LTRB:
            pass
        else:
            raise ValueError(f"Unknown BoxFormat '{box_format}'.")

        xmin, ymin, xmax, ymax = coords

        return cls(
            label=label,
            xmin=xmin,
            ymin=ymin,
            xmax=xmax,
            ymax=ymax,
            confidence=confidence,
        )

    @staticmethod
    def from_txt(
        string: str,
        *,
        box_format=BoxFormat.LTRB,
        relative=False,
        image_size: Optional["tuple[int, int]"] = None,
        separator: Optional[str] = None,
        conf_last: bool = False,
    ) -> "BoundingBox":
        values = string.strip().split(separator)

        if len(values) == 5:
            label, *coords = values
            confidence = None
        elif len(values) == 6:
            if conf_last:
                label, *coords, confidence = values
            else:
                label, confidence, *coords = values
        else:
            raise ParsingError("Syntax error in txt annotation file.")

        try:
            coords = tuple(float(c) for c in coords)
            if confidence is not None:
                confidence = float(confidence)
        except ValueError:
            raise ParsingError("Syntax error in txt annotation file.")

        return BoundingBox.create(
            label=label,
            coords=coords,
            confidence=confidence,
            box_format=box_format,
            relative=relative,
            image_size=image_size,
        )

    @staticmethod
    def from_yolo(
        string: str, *, image_size: "tuple[int, int]", conf_last: bool = False
    ) -> "BoundingBox":
        return BoundingBox.from_txt(
            string,
            box_format=BoxFormat.XYWH,
            relative=True,
            image_size=image_size,
            separator=None,
            conf_last=conf_last,
        )

    @staticmethod
    def from_yolo_darknet(
        string: str,
        *,
        image_size: "tuple[int, int]",
    ) -> "BoundingBox":
        return BoundingBox.from_yolo(string, image_size=image_size, conf_last=False)

    @staticmethod
    def from_yolo_v5(
        string: str,
        *,
        image_size: "tuple[int, int]",
    ) -> "BoundingBox":
        return BoundingBox.from_yolo(string, image_size=image_size, conf_last=True)

    @staticmethod
    def from_yolo_v7(
        string: str,
        *,
        image_size: "tuple[int, int]",
    ) -> "BoundingBox":
        return BoundingBox.from_yolo_v5(string, image_size=image_size)

    @staticmethod
    def from_xml(node: et.Element) -> "BoundingBox":
        label = node.findtext("name")
        box_node = node.find("bndbox")

        if label is None or box_node is None:
            raise ValueError("Syntax error in imagenet annotation format")

        l, t, r, b = (
            box_node.findtext("xmin"),
            box_node.findtext("ymin"),
            box_node.findtext("xmax"),
            box_node.findtext("ymax"),
        )

        if (l is None) or (t is None) or (r is None) or (b is None):
            raise ValueError("Syntax error in imagenet annotation format")

        try:
            coords = tuple(float(c) for c in (l, t, r, b))
        except ValueError:
            raise ParsingError("Syntax error in imagenet annotation format")

        return BoundingBox.create(label=label, coords=tuple(coords))

    @staticmethod
    def from_labelme(node: dict) -> "BoundingBox":
        try:
            label = str(node["label"])
            xs, ys = zip(*node["points"])
            xmin, ymin = min(xs), min(ys)
            xmax, ymax = max(xs), max(ys)
            coords = tuple(float(c) for c in (xmin, ymin, xmax, ymax))
        except (ValueError, KeyError):
            raise ParsingError("Syntax error in labelme annotation file.")

        return BoundingBox.create(label=label, coords=coords)

    @staticmethod
    def from_cvat(node: et.Element) -> "BoundingBox":
        label = node.get("label")
        l, t, r, b = node.get("xtl"), node.get("ytl"), node.get("xbr"), node.get("ybr")

        if (label is None) or (l is None) or (t is None) or (r is None) or (b is None):
            raise ParsingError("Syntax error in CVAT annotation file.")

        try:
            coords = tuple(float(c) for c in (l, t, r, b))
        except ValueError:
            raise ParsingError("Syntax error in CVAT annotation file.")

        return BoundingBox.create(label=label, coords=coords)

    @staticmethod
    def from_via_json(
        region: dict, *, label_key: str = "label_id", confidence_key: str = "confidence"
    ) -> "BoundingBox":
        try:
            region_attrs = region["region_attributes"]
            label = region_attrs[label_key]
            confidence = region_attrs.get(confidence_key)

            shape_attrs = region["shape_attributes"]
            xmin, ymin = shape_attrs["x"], shape_attrs["y"]
            width, height = shape_attrs["width"], shape_attrs["height"]
        except KeyError:
            raise ParsingError("Syntax error in VIA JSON annotation file.")

        return BoundingBox.create(
            label=label,
            coords=(xmin, ymin, width, height),
            confidence=confidence,
            box_format=BoxFormat.LTWH,
        )

    @staticmethod
    def from_yolo_seg(
        string: str,
        *,
        image_size: "tuple[int, int]",
    ) -> "BoundingBox":
        values = string.strip().split()

        if len(values) < 7:
            raise ParsingError(
                "Syntax error in yolo_seg annotation file. There should be at least 7 values."
            )
        elif len(values) % 2 != 1:
            raise ParsingError(
                "Syntax error in yolo_seg annotation file. There should be an odd number of values."
            )
        else:
            label = str(values[0])
            coords_x = tuple(float(value) for value in values[1::2])
            coords_y = tuple(float(value) for value in values[2::2])
            coords = min(coords_x), min(coords_y), max(coords_x), max(coords_y)

        return BoundingBox.create(
            label=label,
            coords=coords,
            box_format=BoxFormat.LTRB,
            relative=True,
            image_size=image_size,
        )

    def to_txt(
        self,
        *,
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        box_format: BoxFormat = BoxFormat.LTRB,
        relative=False,
        image_size: Optional["tuple[int, int]"] = None,
        separator: str = " ",
        conf_last: bool = False,
    ) -> str:
        assert (
            "\n" not in separator
        ), "The newline character '\\n' cannot be used as the separator character."

        if box_format is BoxFormat.LTRB:
            coords = self.ltrb
        elif box_format is BoxFormat.XYWH:
            coords = self.xywh
        elif box_format is BoxFormat.LTWH:
            coords = self.ltwh
        else:
            raise ValueError(f"Unknown BoxFormat '{box_format}'")

        if relative:
            assert (
                image_size is not None
            ), "For relative coordinates, `image_size` should be provided."
            coords = BoundingBox.abs_to_rel(coords, image_size)

        label = self.label
        if label_to_id is not None:
            label = label_to_id[label]

        if isinstance(label, str):
            assert separator not in label, (
                f"The box label '{label}' contains the character '{separator}' which is the same "
                "as the separtor character used for BoundingBox representation in TXT/YOLO format. "
                "This will corrupt the saved annotation file and likely make it unreadable. "
                "Use another character in the label name or `label_to_id` mapping, e.g. use and "
                "underscore instead of a whitespace."
            )

        if self.is_ground_truth:
            line = (label, *coords)
        elif conf_last:
            line = (label, *coords, self._confidence)
        else:
            line = (label, self._confidence, *coords)

        return separator.join(f"{v}" for v in line)

    def to_yolo(
        self,
        *,
        image_size: "tuple[int, int]",
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
        conf_last: bool = False,
    ) -> str:
        return self.to_txt(
            label_to_id=label_to_id,
            box_format=BoxFormat.XYWH,
            relative=True,
            image_size=image_size,
            separator=" ",
            conf_last=conf_last,
        )

    def to_yolo_darknet(
        self,
        *,
        image_size: "tuple[int, int]",
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
    ) -> str:
        return self.to_yolo(
            image_size=image_size, label_to_id=label_to_id, conf_last=False
        )

    def to_yolo_v5(
        self,
        *,
        image_size: "tuple[int, int]",
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
    ) -> str:
        return self.to_yolo(
            image_size=image_size, label_to_id=label_to_id, conf_last=True
        )

    def to_yolo_v7(
        self,
        *,
        image_size: "tuple[int, int]",
        label_to_id: Optional[Mapping[str, Union[int, str]]] = None,
    ) -> str:
        return self.to_yolo_v5(image_size=image_size, label_to_id=label_to_id)

    def to_labelme(self) -> dict:
        xmin, ymin, xmax, ymax = self.ltrb

        return {
            "label": self.label,
            "points": [[xmin, ymin], [xmax, ymax]],
            "shape_type": "rectangle",
        }

    def to_xml(self) -> et.Element:
        obj_node = et.Element("object")
        et.SubElement(obj_node, "name").text = self.label
        box_node = et.SubElement(obj_node, "bndbox")

        for tag, coord in zip(("xmin", "ymin", "xmax", "ymax"), self.ltrb):
            et.SubElement(box_node, tag).text = f"{coord}"

        return obj_node

    def to_cvat(self) -> et.Element:
        xtl, ytl, xbr, ybr = self.ltrb
        return et.Element(
            "box",
            attrib={
                "label": self.label,
                "xtl": f"{xtl}",
                "ytl": f"{ytl}",
                "xbr": f"{xbr}",
                "ybr": f"{ybr}",
            },
        )

    def to_via_json(
        self, *, label_key: str = "label_id", confidence_key: str = "confidence"
    ) -> dict:
        assert (
            label_key != confidence_key
        ), f"Label key '{label_key}' and confidence key '{confidence_key}' should be different."

        shape_attributes = {
            "name": "rect",
            "x": self.xmin,
            "y": self.ymin,
            "width": self.width,
            "height": self.height,
        }

        region_attributes: dict[str, Any] = {label_key: self.label}

        if self.confidence is not None:
            region_attributes[confidence_key] = self.confidence

        return {
            "shape_attributes": shape_attributes,
            "region_attributes": region_attributes,
        }

    def __eq__(self, other):
        if not isinstance(other, BoundingBox):
            raise NotImplementedError

        return (
            self.label == other.label
            and self.xmin == other.xmin
            and self.ymin == other.ymin
            and self.xmax == other.xmax
            and self.ymax == other.ymax
            and self.confidence == other.confidence
        )

    def __repr__(self) -> str:
        return (
            f"BoundingBox(label: {self.label}, xmin: {self._xmin}, ymin: {self._ymin}, "
            f"xmax: {self._xmax}, ymax: {self._ymax}, confidence: {self._confidence})"
        )



from collections import defaultdict
from itertools import groupby
from typing import Callable, Hashable, Iterable, TypeVar

U = TypeVar("U")
V = TypeVar("V", bound=Hashable)


def grouping(it: Iterable[U], by_key: Callable[[U], V]) -> "dict[V, list[U]]":
    result = defaultdict(list)
    for item in it:
        result[by_key(item)].append(item)
    return result


def all_equal(iterable: Iterable) -> bool:
    """https://stackoverflow.com/a/3844948/6324055"""
    g = groupby(iterable)
    return next(g, True) and not next(g, False)


def mean(it: Iterable[float]) -> float:
    sum_ = 0.0
    count = 0

    for value in it:
        sum_ += value
        count += 1

    if count == 0:
        return float("nan")

    return sum_ / count
